from typing import Callable
from typing import Self

import numpy as np

from .cov import cov_to_corr
from .utils import array_like


def get_grid_points(coords: list[np.ndarray]) -> np.ndarray:
    return np.vstack([C.ravel() for C in np.meshgrid(*coords, indexing="ij")]).T


def squeeze_coords(coords: list[np.ndarray]) -> np.ndarray:
    if len(coords) == 1:
        return coords[0]
    return coords


def edges_to_coords_1d(edges: np.ndarray) -> np.ndarray:
    return 0.5 * (edges[:-1] + edges[1:])


def coords_to_edges_1d(coords: np.ndarray) -> np.ndarray:
    delta = np.diff(coords)[0]
    return np.hstack([coords - 0.5 * delta, [coords[-1] + 0.5 * delta]])


def edges_to_coords(
    edges: np.ndarray | list[np.ndarray],
) -> np.ndarray | list[np.ndarray]:
    """Compute bin center coordinates from evenly spaced bin edges."""
    coords = None
    if np.isscalar(edges[0]):
        coords = edges_to_coords_1d(edges)
    else:
        coords = [edges_to_coords_1d(e) for e in edges]
    return coords


def coords_to_edges(coords: np.ndarray | list[np.ndarray]) -> np.ndarray:
    """Compute bin edges from evenly spaced bin coordinates."""
    edges = None
    if np.isscalar(coords[0]):
        edges = coords_to_edges_1d(coords)
    else:
        edges = [coords_to_edges_1d(c) for c in coords]
    return edges


class Grid:
    def __init__(
        self, coords: list[np.ndarray] = None, edges: list[np.ndarray] = None
    ) -> None:
        self.coords = coords
        self.edges = edges

        if (self.coords is None) and (self.edges is not None):
            self.coords = edges_to_coords(self.edges)

        if (self.edges is None) and (self.coords is not None):
            self.edges = coords_to_edges(self.coords)

        self.shape = tuple([len(c) for c in self.coords])
        self.ndim = len(self.shape)
        self.size = np.prod(self.shape)
        self.cell_volume = np.prod([c[1] - c[0] for c in self.coords])

        self.limits = [(e[0], e[-1]) for e in self.edges]
        self.limits = np.array(self.limits)

        self.ranges = [(xmax - xmin) for (xmin, xmax) in self.limits]

    def points(self) -> np.ndarray:
        return get_grid_points(self.coords)

    def meshgrid(self) -> np.ndarray:
        return np.meshgrid(*self.coords, indexing="ij")

    def normalize(self) -> None:
        values_sum = np.sum(self.values)
        if values_sum > 0.0:
            self.values = self.values / values_sum / self.cell_volume

    def grid_points(self) -> np.ndarray:
        return get_grid_points(self.coords)


class Histogram(Grid):
    def __init__(
        self,
        values: np.ndarray = None,
        coords: np.ndarray = None,
        edges: np.ndarray = None,
    ) -> None:
        if (values is not None):
            if (values.ndim == 1):
                return Histogram1D(values=values, coords=coords, edges=edges)

        super().__init__(coords=coords, edges=edges)

        self.values = values
        if self.values is None:
            self.values = np.zeros(self.shape)

    def copy(self) -> Self:
        return Histogram(values=self.values, edges=self.edges)

    def max_indices(self) -> tuple[np.ndarray]:
        return np.unravel_index(np.argmax(self.values), self.shape)

    def scale_max(self) -> None:
        max_value = np.max(self.values)
        if max_value > 0.0:
            self.values = self.values / max_value

    def sample(self, size: int, noise: float = 0.0) -> np.ndarray:
        return sample_hist(self, size=size, noise=noise)

    def project(self, axis: int | tuple[int, ...]) -> Self:
        return project(self, axis, squeeze=True)

    def slice(
        self,
        axis: int | tuple[int, ...],
        ind: int | tuple[int, ...] | list[tuple[int, ...]],
        return_indices: bool = False,
    ) -> Self | tuple[Self, list[slice]]:
        return slice_(self, axis=axis, ind=ind, return_indices=return_indices)

    def slice_ellipsoid(
        self,
        axis: tuple[int, ...],
        rmin: float,
        rmax: float,
        return_indices: bool = False,
    ) -> Self | tuple[Self, list[slice]]:
        return slice_ellipsoid(self, axis=axis, rmin=rmin, rmax=rmax, return_indices=return_indices)

    def slice_contour(
        self,
        axis: tuple[int, ...],
        rmin: float,
        rmax: float,
        return_indices: bool = False,
    ) -> Self | tuple[Self, list[slice]]:
        return slice_contour(self, axis=axis, rmin=rmin, rmax=rmax, return_indices=return_indices)

    def bin(self, points: np.ndarray, density: bool = True) -> np.ndarray:
        self.values, _ = np.histogramdd(points, bins=self.edges, density=density)
        return np.copy(self.values)

    def __call__(self, **kwargs) -> np.ndarray:
        return self.bin(**kwargs)


class SparseHistogram(Grid):
    def __init__(
        self,
        values: np.ndarray,
        indices: np.ndarray,
        coords: np.ndarray = None,
        edges: np.ndarray = None,
    ) -> None:
        """Constructor.

        Parameters
        ----------
        values: ndarray, shape (K,)
            Counts in each bin. Does not need to be normalized.
        indices: ndarray, shape (K, N)
            Indices of nonzero cells in N-dimensional hist.
        """
        super().__init__(coords=coords, edges=edges)
        self.values = values
        self.indices = indices
        self.size = len(self.values)  # number of nonzero cells

    def sample(self, size: int, noise: float = 0.0) -> np.ndarray:
        return sample_sparse_hist(self, size=size, noise=noise)


class Histogram1D:
    def __init__(
        self,
        values: np.ndarray = None,
        coords: np.ndarray = None,
        edges: np.ndarray = None,
    ) -> None:
        self.coords = coords
        self.edges = edges

        if (self.coords is None) and (self.edges is not None):
            self.coords = edges_to_coords(self.edges)

        if (self.edges is None) and (self.coords is not None):
            self.edges = coords_to_edges(self.coords)

        self.shape = (len(self.coords),)
        self.ndim = 1
        self.size = len(self.coords)
        self.cell_volume = self.coords[1] - self.coords[0]
        self.limits = np.array((self.edges[0], self.edges[-1]))
        self.range = self.limits[1] - self.limits[0]

        self.values = values
        if self.values is None:
            self.values = np.zeros(self.shape)

    def copy(self) -> Self:
        return Histogram1D(values=self.values, edges=self.edges)

    def normalize(self) -> None:
        values_sum = np.sum(self.values)
        if values_sum > 0.0:
            self.values = self.values / values_sum / self.cell_volume

    def bin(self, points: np.ndarray, density: bool = True) -> np.ndarray:
        self.values, _ = np.histogram(points, bins=self.edges, density=density)
        return np.copy(self.values)

    def __call__(self, **kwargs) -> np.ndarray:
        return self.bin(**kwargs)


def slice_idx(
    ndim: int,
    axis: int | tuple[int, ...],
    ind: int | tuple[int, ...] | list[tuple[int, ...]],
) -> tuple[slice, ...]:
    """Return planar slice index array.

    Parameters
    ----------
    ndim: int
        The number of elements in the slice index array. (The number of dimensions in the array to be sliced.)
    axis: int or tuple[int, ...]
        The sliced axes.
    ind: int, tuple[int, ...] or list[tuple[int, ...]]
        The indices along the sliced axes. If a tuple is provided, this defines the (min, max) index.

    Returns
    -------
    idx: tuple
        The slice index array. A slice of the array `values` may then be accessed as `values[idx]`.
    """
    # Make list if only one axis provided.
    if type(axis) is int:
        axis = [axis]
        # Can also provide only one axis but provide a tuple (selects range of 
        # indices) or list (selects specific indices) for ind.
        if not (type(ind) is int):
            ind = [ind]

    # Make list if only one ind provided.
    if type(ind) is int:
        ind = [ind]

    # Initialize the slice index to select all elements.
    idx = ndim * [slice(None)]

    # If any indices were provided, add them to `idx`.
    for k, item in zip(axis, ind):
        if item is None:
            continue
        elif (type(item) is tuple) and (len(item) == 2):
            idx[k] = slice(item[0], item[1])
        else:
            # Could be int or list of ints
            idx[k] = item

    return tuple(idx)


def slice_(
    hist: Histogram,
    axis: int | tuple[int, ...],
    ind: int | tuple[int, ...] | list[tuple[int, ...]],
    return_indices: bool = False,
) -> Histogram | tuple[Histogram, list[slice]]:

    # Get slice idx
    ndim = hist.ndim
    idx = slice_idx(ndim=hist.ndim, axis=axis, ind=ind)

    # Slice the array
    values_new = hist.values[idx]
    values_new = np.squeeze(values_new)

    # Get remaining array coordinates
    coords_new = []
    for axis in range(ndim):
        if type(idx[axis]) is slice:            
            # Selects indices using slice object.
            ilo = idx[axis].start
            if ilo is None:
                ilo = 0
    
            ihi = idx[axis].stop
            if ihi is None:
                ihi = hist.shape[axis]
    
            if ihi - ilo > 1:
                coords_new.append(hist.coords[axis][idx[axis]])
            
        elif (type(idx[axis]) is tuple) and (len(idx[axis]) == 2):
            # Selects indices using a tuple of ints (imin, imax).
            ilo, ihi = idx[axis]

        elif array_like(idx[axis]) and (len(idx[axis]) > 1):
            # Selects indices using a list or ndarray.
            coords_new.append(hist.coords[axis][list(idx[axis])])

        else:
            # Selects a single index; this axis is not in the sliced array.
            continue
    
    # Return new histogram
    hist_new = Histogram(values=values_new, coords=coords_new)

    if return_indices:
        return hist_new, idx
    return hist_new


def project_values(values: np.ndarray, axis: int | tuple[int, ...]) -> np.ndarray:
    if values.ndim == 1:
        return hist

    # Sum over axes
    if type(axis) is int:
        axis = [axis]
    axis = tuple(axis)
    axis_sum = tuple([i for i in range(values.ndim) if i not in axis])

    values_proj = None
    if len(axis) != values.ndim:
        values_proj = np.sum(values, axis_sum)
    else:
        values_proj = values

    # Order the remaining axes
    loc = list(range(values_proj.ndim))
    destination = np.zeros(values_proj.ndim, dtype=int)
    for i, index in enumerate(np.argsort(axis)):
        destination[index] = i
    for i in range(values_proj.ndim):
        if loc[i] != destination[i]:
            j = loc.index(destination[i])
            values_proj = np.swapaxes(values_proj, i, j)
            (loc[i], loc[j]) = (loc[j], loc[i])

    return values_proj


def project(
    hist: Histogram, axis: int | tuple[int, ...], squeeze: bool = True
) -> Histogram | Histogram1D:
    
    values_proj = project_values(hist.values, axis=axis)
    if (values_proj.ndim == 1) and squeeze:
        axis = int(np.squeeze(axis))
        coords_proj = hist.coords[axis]
        hist_proj = Histogram1D(values_proj, coords_proj)
    else:
        coords_proj = [hist.coords[i] for i in axis]
        hist_proj = Histogram(values_proj, coords_proj)
    return hist_proj


def project_contour_slice_1d(
    hist: Histogram,
    axis: int = 0,
    lmin: float = 0.0,
    lmax: float = 1.0,
    hist_proj: Histogram = None,
) -> Histogram:
    """Apply contour slice in N- 1 dimensions, then project onto the remaining dimension.

    Parameters
    ----------
    hist: Histogram
        Distribution on N-dimensional hist.
    axis: int
        The projection axis.
    lmin, lmax: float
        Min and max contour levels of the (N-1)-dimensional projection of `values`,
        normalized the range [0, 1].
    hist_proj: Histogram
        The (N-1)-dimensional projection of `hist` onto all dimensions other than `axis`.
        If not provided, it will be computed within the function.

        Shape:
        ```
        tuple([hist.shape[i] for i in range(hist.ndim) if i != axis])
        ```

    Returns
    -------
    hist
        The 1D projection of the (N-1)D slice.
    """
    axis_proj = [i for i in range(hist.ndim) if i != axis]
    
    if hist_proj is None:
        hist_proj = hist.project(axis_proj)
        
    hist_proj.values = hist_proj.values / np.max(hist_proj.values)

    idx = slice_idx(
        ndim=hist.ndim,
        axis=axis_proj,
        ind=np.where(
            np.logical_and(
                hist_proj.values >= lmin, 
                hist_proj.values <= lmax,
            )
        ),
    )

    # `values[idx]` will give a two-dimensional array. Normally we need to sum over
    # the first axis. If `axis == 0`, we need to sum over the second axis.
    values_proj_1d = np.sum(hist.values[idx], axis=int(axis == 0))
    coords_proj_1d = hist.coords[axis]
    return Histogram1D(values=values_proj_1d, coords=coords_proj_1d)


def project_contour_slice_2d(
    hist: Histogram,
    axis: tuple[int, int] = (0, 1),
    lmin: float = 0.0,
    lmax: float = 1.0,
    hist_proj: Histogram = None,
) -> np.ndarray:
    """Apply contour slice in n - 2 dimensions, then project onto the remaining two dimensions.

    Parameters
    ----------
    hist: Histogram
        Distribution on N-dimensional hist.
    axis: tuple
        The 2D projection axis.
    lmin, lmax: float
        Min and max contour levels of the (n-2)-dimensional projection of `values`,
        normalized the the range [0, 1].
    values_proj: ndarray, shape = [values.shape[i] for i in range(values.ndim) if i != axis]
        The (n-2)-dimensional projection of `values` onto all dimensions other than `axis`.
        (If not provided, it will be computed within the function.)

    Returns
    -------
    hist
        The 2D projection of the (N-2)D slice.
    """
    axis_proj = [k for k in range(hist.ndim) if k not in axis]
    axis_proj = tuple(axis_proj)

    if hist_proj is None:
        hist_proj = project(hist, axis=axis_proj)

    hist_proj.scale_max()

    idx = slice_idx(
        ndim=hist.ndim,
        axis=axis_proj,
        ind=np.where(
            np.logical_and(
                hist_proj.values >= lmin,
                hist_proj.values <= lmax
            )
        )
    )

    # `values[idx]` will give a three-dimensional array. Normally we need to sum over
    # the first axis. If `axis == (0, 1)`, we need to sum over the third axis.
    # If `axis == (0, n - 1), we need to sum over the second axis.
    _axis_proj = (1, 2)
    if axis == (0, 1):
        _axis_proj = (0, 1)
    elif axis == (0, hist.ndim - 1):
        _axis_proj = (0, 2)

    # Two elements of `idx` will be `slice(None)`; these are the elements in `axis`.
    # These will always be in order. So, if `axis[0] > axis[1]`, we need to flip
    # `axis_proj`. Need a way to handle this automatically.
    if axis[0] > axis[1]:
        _axis_proj = tuple(reversed(_axis_proj))

    values_proj = project_values(hist.values[idx], axis=_axis_proj)
    coords_proj = [hist.coords[_axis] for _axis in _axis_proj]
    return Histogram(values_proj, coords_proj)


def copy_values_into_new_dim(
    values: np.ndarray,
    shape: tuple[int, ...],
    axis: int = -1,
    method: str = "broadcast",
    copy: bool = False,
) -> np.ndarray:
    """Copy hist into one or more new dimensions.

    https://stackoverflow.com/questions/32171917/how-to-copy-a-2d-array-into-a-3rd-dimension-n-times

    Parameters
    ----------
    values: ndarray
        An n-dimensional hist.
    shape: d-tuple of ints
        The shape of the new dimensions.
    axis: int (0 or -1)
        If 0, the new dimensions will be inserted before the first axis. If -1,
        the new dimensions will be inserted after the last axis. I think
        values other than 0 or -1 should work; this does not currently
        work, at least for `method='broadcast'`, last I checked.
    method: {'repeat', 'broadcast'}
        Whether to use `np.repeat` or `np.expand_dims` and `np.broadcast_to`. The
        'broadcast' method is faster.
    """
    if not array_like(shape):
        shape = (shape,)

    if method == "repeat":
        for i in range(len(shape)):
            values = np.repeat(np.expand_dims(values, axis), shape[i], axis=axis)
        return values
    elif method == "broadcast":
        if axis == 0:
            new_shape = shape + values.shape
        elif axis == -1:
            new_shape = values.shape + shape
        else:
            raise ValueError("Cannot yet handle axis != 0, -1.")

        for _ in range(len(shape)):
            values = np.expand_dims(values, axis)

        if copy:
            return np.broadcast_to(values, new_shape).copy()
        else:
            return np.broadcast_to(values, new_shape)
    else:
        raise ValueError


def mean(hist: Histogram) -> np.ndarray:
    x = [np.average(C, weights=hist.values) for C in hist.meshgrid()]
    x = np.array(x)
    return x


def cov(hist: Histogram) -> np.ndarray:
    def cov_2d(values: np.ndarray, coords: list[np.ndarray]) -> np.ndarray:
        COORDS = np.meshgrid(*coords, indexing="ij")
        S = np.zeros((values.ndim, values.ndim))
        values_sum = np.sum(values)
        if values_sum > 0:
            mean = np.array([np.average(C, weights=values) for C in COORDS])
            for i in range(values.ndim):
                for j in range(i + 1):
                    X = COORDS[i] - mean[i]
                    Y = COORDS[j] - mean[j]
                    EX = np.sum(values * X) / values_sum
                    EY = np.sum(values * Y) / values_sum
                    EXY = np.sum(values * X * Y) / values_sum
                    S[i, j] = S[j, i] = EXY - EX * EY
        return S

    ndim = hist.ndim
    if ndim < 3:
        return cov_2d(hist.values, hist.coords)

    S = np.zeros((ndim, ndim))
    for i in range(ndim):
        for j in range(i):
            # Compute 2 x 2 covariance matrix from this projection.
            axis = (i, j)
            hist_proj = project(hist, axis)
            S_proj = cov_2d(hist_proj.values, hist_proj.coords)

            # Update elements of n x n covariance matrix. This will update
            # some elements multiple times, but it should not matter.
            S[i, i] = S_proj[0, 0]
            S[j, j] = S_proj[1, 1]
            S[i, j] = S[j, i] = S_proj[0, 1]
    return S


def corr(hist: Histogram) -> np.ndarray:
    return cov_to_corr(cov(hist))


def ellipsoid_radii(
    coords: list[np.ndarray], covariance_matrix: np.ndarray
) -> np.ndarray:
    """Return covariance ellipsoid radii from grid coordinates and covariance matrix.

    Radius is defined as x^T Sigma^-1^T. This function computes the radius
    at every point on the hist.

    This is quite slow when n > 4.

    Parameters
    ----------
    coords: list[np.ndarray]
        Grid coordinates along each dimension.
    covariance_matrix: np.ndarray
        N x N covariance matrix.

    Returns
    -------
    ndarray: np.ndarray
        Radius x^T Sigma^-1^T x at each point in hist.
    """
    shape = tuple([len(c) for c in coords])
    COORDS = np.meshgrid(*coords, indexing="ij")

    S = covariance_matrix
    S_inv = np.linalg.inv(S)

    R = np.zeros(shape)
    for ii in np.ndindex(shape):
        v = np.array([C[ii] for C in COORDS])
        R[ii] = np.sqrt(np.linalg.multi_dot([v.T, S_inv, v]))
    return R


def radial_density(
    hist: Histogram,
    radii: np.ndarray,
    dr: float = None,
) -> np.ndarray:
    """Return average density within ellipsoidal shells.

    Parameters
    ----------
    hist: ndarray
        An n-dimensional hist.
    R: ndarray, same shape as `values`.
        Gives the radius at each point in `values`.
    radii: ndarray, shape (k,)
        Radii at which to evaluate the density.
    dr: float
        The radial shell width.

    Returns
    -------
    ndarray, shape (k,)
        The average density within each ellipsoidal shell.
    """
    values = hist.values
    covariance_matrix = cov(hist)
    R = ellipsoid_radii(hist.coords, covariance_matrix)
    if dr is None:
        dr = 0.5 * np.max(R) / (len(R) - 1)

    values_r = []
    for r in radii:
        values_masked = np.ma.masked_where(np.logical_or(R < r, R > r + dr), values)
        values_r.append(np.mean(values_masked))
    return np.array(values_r)


def blur_hist(hist: Histogram, sigma: float) -> np.ndarray:
    """Call scipy.ndhist.gaussian_filter."""
    hist.values = ndhist.gaussian_filter(hist.values, sigma)
    return hist


def clip_hist(
    hist: Histogram, lmin: float = None, lmax: float = None, frac: bool = False
) -> np.ndarray:
    """Clip between lmin and lmax, can be fractions or absolute values."""
    if not (lmin or lmax):
        return hist
    if frac:
        values = hist.values
        values_max = np.max(values)
        if lmin:
            lmin = values_max * lmin
        if lmax:
            lmax = values_max * lmax
    hist.values = np.clip(values, lmin, lmax)
    return hist


def fill_hist(hist: np.ndarray, fill_value: float = None) -> np.ndarray:
    """Call numpy.ma.filled."""
    hist.values = np.ma.filled(hist.values, fill_value=fill_value)
    return hist


def thresh_hist(hist: Histogram, lmin: float = None, frac: bool = False) -> np.ndarray:
    if lmin:
        if frac:
            lmin = lmin * hist.values.max()
        hist.values[hist.values < lmin] = 0.0
    return hist


def sample_hist(hist: Histogram, size: int = 100, noise: float = 0.0) -> np.ndarray:

    values = hist.values
    edges = hist.edges

    idx = np.flatnonzero(values)
    pdf = values.ravel()[idx]
    pdf = pdf / np.sum(pdf)
    idx = np.random.choice(idx, size, replace=True, p=pdf)
    idx = np.unravel_index(idx, shape=values.shape)
    lb = [edges[axis][idx[axis]] for axis in range(values.ndim)]
    ub = [edges[axis][idx[axis] + 1] for axis in range(values.ndim)]

    points = np.squeeze(np.random.uniform(lb, ub).T)
    if noise:
        for axis in range(points.shape[1]):
            delta = ub[axis] - lb[axis]
            points[:, axis] += (
                noise * 0.5 * np.random.uniform(-delta, delta, size=points.shape[0])
            )
    return points


def sample_sparse_hist(
    hist: SparseHistogram, size: int = 100, noise: float = 0.0
) -> np.ndarray:
    """Sample from sparse histogram.

    Parameters
    ----------
    hist: SparseHist
        Sparse N-dimensional hist.
    size: int
        The number of samples to draw.
    noise: float
        Add noise to each particle; a number is drawn uniformly from a box centered on the
        particle with dimensions equal to the histogram bin dimensions. `noise` scales the
        box dimensions relative to the bin dimensions.

    Returns
    -------
    ndarray, shape (size, N)
        Samples drawn from the distribution.
    """
    values = hist.values
    indices = hist.indices
    edges = hist.edges

    shape = [len(e) - 1 for e in edges]
    indices_flat = np.ravel_multi_index(indices.T, shape)
    idx = np.random.choice(
        indices_flat, size=size, replace=True, p=(values / np.sum(values))
    )
    idx = np.unravel_index(idx, shape=shape)
    lb = [edges[axis][idx[axis]] for axis in range(len(shape))]
    ub = [edges[axis][idx[axis] + 1] for axis in range(len(shape))]
    points = np.squeeze(np.random.uniform(lb, ub).T)
    if noise:
        delta = 0.5 * noise * np.array([np.mean(np.diff(e)) for e in edges])
        points += np.random.uniform(low=-delta, high=delta, size=points.shape)
    return points



import pytest
import numpy as np

import psdist as ps


def test_rms_ellipsoid_volume_identity():
    S = np.eye(2)
    assert np.isclose(ps.cov.rms_ellipsoid_volume(S), 1.0)


def test_rms_ellipsoid_volume_diagonal():
    S = np.diag([2.0, 3.0])
    assert np.isclose(ps.cov.rms_ellipsoid_volume(S), np.sqrt(6.0))


def test_projected_emittances_identity():
    S = np.eye(4)
    assert np.allclose(ps.cov.projected_emittances(S), [1.0, 1.0])


def test_intrinsic_emittances_identity():
    S = np.eye(4)
    assert np.allclose(ps.cov.intrinsic_emittances(S), [1.0, 1.0])


def test_twiss_identity():
    S = np.eye(4)
    twiss_params = ps.cov.twiss(S)
    assert np.allclose(twiss_params, [(0.0, 1.0, 1.0), (0.0, 1.0, 1.0)])


def test_unit_symplectic_matrix():
    U = ps.cov.unit_symplectic_matrix(4)
    expected = np.array([[0, 1, 0, 0], [-1, 0, 0, 0], [0, 0, 0, 1], [0, 0, -1, 0]])
    assert np.allclose(U, expected)


def test_normalization_matrix_from_twiss_2d():
    alpha = 0.0
    beta = 1.0
    emittance = 1.0
    V_inv = ps.cov.normalization_matrix_from_twiss_2d(alpha, beta, emittance)
    expected = np.eye(2)
    assert np.allclose(V_inv, expected)


def test_cov_to_corr_identity():
    S = np.eye(2)
    assert np.allclose(ps.cov.cov_to_corr(S), S)


def test_rms_ellipse_params():
    S = np.eye(2)
    cx, cy, theta = ps.cov.rms_ellipse_params(S)
    assert np.isclose(cx, 1.0)
    assert np.isclose(cy, 1.0)
    assert np.isclose(theta, 0.0)



import numpy as np
import seaborn as sns
from matplotlib import patches
from matplotlib import pyplot as plt

from ..cov import rms_ellipse_params as _rms_ellipse_params


def plot_ellipse(
    r1: float = 1.0,
    r2: float = 1.0,
    angle: float = 0.0,
    center: tuple[float, float] = None,
    ax=None,
    **kws,
):
    """Plot ellipse with semi-axes `c1`,`c2` tilted `angle`radians below the x axis."""
    kws.setdefault("fill", False)
    kws.setdefault("color", "black")
    kws.setdefault("lw", 1.25)

    if center is None:
        center = (0.0, 0.0)

    d1 = r1 * 2.0
    d2 = r2 * 2.0
    angle = -np.degrees(angle)
    ax.add_patch(patches.Ellipse(center, d1, d2, angle=angle, **kws))
    return ax


def plot_circle(
    r: float = 1.0, center: tuple[float, float] = None, ax=None, **kws
):
    """Plot circle of radius r."""
    if center is None:
        center = (0.0, 0.0)
    return plot_ellipse(r, r, center=center, ax=ax, **kws)



def plot_rms_ellipse_cov(
    cov_matrix: np.ndarray,
    center: np.ndarray = None,
    level: float = 1.0,
    ax=None,
    **ellipse_kws,
):
    """Plot RMS ellipse from 2 x 2 covariance matrix."""
    if center is None:
        center = (0.0, 0.0)
    if type(level) not in [list, tuple, np.ndarray]:
        level = [level]
    r1, r2, angle = _rms_ellipse_params(cov_matrix)
    for level in level:
        plot_ellipse(
            r1 * level, r2 * level, angle=angle, center=center, ax=ax, **ellipse_kws
        )
    return ax


def cubehelix_cmap(color: str = "red", dark: float = 0.20):
    kws = dict(
        n_colors=12,
        rot=0.0,
        gamma=1.0,
        hue=1.0,
        light=1.0,
        dark=dark,
        as_cmap=True,
    )

    cmap = None
    if color == "red":
        cmap = sns.cubehelix_palette(start=0.9, **kws)
    elif color == "pink":
        cmap = sns.cubehelix_palette(start=0.8, **kws)
    elif color == "blue":
        cmap = sns.cubehelix_palette(start=2.8, **kws)
    else:
        raise ValueError
    return cmap



    



import numpy as np

from .utils import sphere_volume


class Distribution:
    def __init__(self, ndim: int) -> None:
        self.ndim = ndim

    def prob(self, x: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    def sample(self, size: int) -> np.ndarray:
        raise NotImplementedError


class Gaussian(Distribution):
    def __init__(self, ndim: int) -> None:
        super().__init__(ndim=ndim)

    def prob(self, x: np.ndarray) -> np.ndarray:
        return np.exp(-0.5 * np.sum(x**2, axis=1)) / np.sqrt(2.0 * np.pi)

    def entropy(self):
        return 0.5 * (np.log(2.0 * np.pi) + 1.0)

    def sample(self, size: int) -> np.ndarray:
        return np.random.normal(size=(size, self.ndim))


class Waterbag(Distribution):
    def __init__(self, ndim: int) -> None:
        super().__init__(ndim=ndim)
        self.r_max = np.sqrt(self.ndim + 2)

    def prob(self, x: np.ndarray) -> np.ndarray:
        r = np.sqrt(np.sum(np.square(x), axis=1))
        prob = np.zeros(x.shape[0])
        prob[r <= self.r_max] = 1.0 / sphere_volume(self.r_max, self.ndim)
        return prob

    def sample(self, size: int) -> np.ndarray:
        x = np.random.normal(size=(size, self.ndim))
        scale = 1.0 / np.sqrt(np.sum(x**2, axis=1))
        x = x * scale[:, None]
        scale = np.random.uniform(0.0, 1.0, size=size) ** (1.0 / self.ndim)
        scale = scale * self.r_max
        x = x * scale[:, None]
        return x


def get_distribution(name: str, ndim: int, *args, **kwargs) -> Distribution:
    constructors = {
        "gaussian": Gaussian,
        "waterbag": Waterbag,
    }
    constructor = constructors[name]
    return constructor(ndim=ndim, *args, **kwargs)



from typing import Any
from typing import Union

import numpy as np
import scipy.special


def array_like(array: Any) -> bool:
    return np.ndim(np.array(array, dtype=object)) > 0


def symmetrize(array: np.ndarray) -> np.ndarray:
    """Return a symmetrized version of array.

    array : A square upper or lower triangular matrix.
    """
    return array + array.T - np.diag(array.diagonal())


def random_choice_no_replacement(array: np.ndarray, size: int) -> np.ndarray:
    """Choose random elements from array without replacement.

    If 0 < size < 1, we select `size * len(array)` elements.
    """
    if type(array) in (list, tuple):
        array = np.array(array)

    if size is None:
        return array

    if size < 0 or size > array.shape[0]:
        raise ValueError("Number of samples must be < number of points.")
    if 0 < size < 1:
        size = size * array.shape[0]
    size = int(size)

    idx = np.random.choice(array.shape[0], size, replace=False)
    return array[idx]


def sphere_surface_area(r: float = 1.0, ndim: int = 3) -> float:
    factor = 2.0 * np.pi ** (0.5 * ndim)
    factor = factor / scipy.special.gamma(0.5 * ndim)
    return factor * (r ** (ndim - 1))


def sphere_volume(r: float = 1.0, ndim: int = 3) -> float:
    factor = (np.pi ** (0.5 * ndim)) / scipy.special.gamma(1.0 + 0.5 * ndim)
    return factor * (r**ndim)


def sphere_shell_volume(rmin: float = 0.0, rmax: float = 1.0, ndim: int = 3):
    return sphere_volume(rmax, ndim=ndim) - sphere_volume(rmin, ndim)


def rotation_matrix(angle: float) -> np.ndarray:
    c, s = np.cos(angle), np.sin(angle)
    return np.array([[c, s], [-s, c]])


# The following three functions allow saving/loading ragged arrays in .npz format.
# This is useful if we have multiple coordinate arrays with a different number of
# points in each array.
# (Source: https://tonysyu.github.io/ragged-arrays.html#.YKVwQy9h3OR)


def stack_ragged(arrays: list[np.ndarray], axis: int = 0) -> tuple[np.ndarray]:
    """Stacks list of arrays along first axis.

    Example: (25, 4) + (75, 4) -> (100, 4). It also returns the indices at
    which to split the stacked array to regain the original list of arrays.
    """
    lengths = [np.shape(a)[axis] for a in arrays]
    idx = np.cumsum(lengths[:-1])
    stacked = np.concatenate(arrays, axis=axis)
    return stacked, idx


def save_stacked_array(path: str, arrays: list[np.ndarray], axis: int = 0) -> None:
    """Save list of ragged arrays as single stacked array. The index from
    `stack_ragged` is also saved."""
    stacked, idx = stack_ragged(arrays, axis=axis)
    np.savez(path, stacked_array=stacked, stacked_index=idx)


def load_stacked_arrays(path: str, axis: int = 0) -> list[np.ndarray]:
    """ "Load stacked ragged array from .npz file as list of arrays."""
    npz_file = np.load(path)
    idx = npz_file["stacked_index"]
    stacked = npz_file["stacked_array"]
    return np.split(stacked, idx, axis=axis)



"""Plotting routines for multi-dimensional images."""
import matplotlib.pyplot as plt
import numpy as np
import scipy.ndimage
import ultraplot as uplt
from ipywidgets import interactive
from ipywidgets import widgets

import psdist as ps
from ..hist import Histogram
from ..hist import Histogram1D
from . import core


def _process_values_fill(values: np.ndarray, fill_value: float) -> np.ndarray:
    return np.ma.filled(values, fill_value=fill_value)


def _process_values_thresh(
    values: np.ndarray, lmin: 0.0, frac: bool = False
) -> np.ndarray:
    if frac:
        lmin = lmin * values.max()
    values[values < lmin] = 0.0
    return values


def _process_values_clip(
    values: np.ndarray, lmin: float = None, lmax: float = None, frac: bool = False
) -> np.ndarray:
    """Clip between lmin and lmax, can be fractions or absolute values."""
    if not (lmin or lmax):
        return values
    if frac:
        f_max = np.max(f)
        if lmin:
            lmin = f_max * lmin
        if lmax:
            lmax = f_max * lmax
    return np.clip(f, lmin, lmax)


def _process_values_blur(values: np.ndarray, sigma: float) -> np.ndarray:
    return scipy.ndimage.gaussian_filter(values, sigma)


def _process_values_scale_max(values: np.ndarray) -> np.ndarray:
    values_max = np.max(values)
    if values_max > 0.0:
        values = values / values_max
    return values


def process_hist(
    hist: Histogram | Histogram1D,
    fill_value: float = None,
    thresh: float = None,
    thresh_type: str = "abs",
    clip: tuple[float] = None,
    clip_type: str = "abs",
    blur: float = None,
    scale: str = None,
) -> Histogram:
    """Return processed image.

    Defaults leave histogram unchanged.

    Parameters
    ----------
    hist : Histogram
        A two-dimensional histogram.
    fill_value : float
        Fill masked elements of `f` with this value.
    thresh : float
        Set elements below this value to zero.
    thresh_type : {"frac", "abs"}
        Whether `thresh` is absolute or relative to peak.
    clip : (lmin, lmax)
        Clip (limit) elements to within the range [lmin, lmax].
    clip_type : {"frac", "abs"}
        Whether `clip` is absolute or relative to peak.
    scale : {None, "max", "density"}
        Normalize by max, volume, or None.
    blur : float
        Kernel width for gaussian filter.
    """
    values = hist.values.copy()

    if fill_value is not None:
        values = _process_values_fill(values, fill_value)
    if thresh is not None:
        values = _process_values_thresh(values, thresh, frac=(thresh_type == "frac"))
    if clip is not None:
        values = _process_values_clip(
            values, lmin=clip[0], lmax=clip[1], frac=(clip_type == "fract")
        )
    if blur is not None:
        values = _process_values_blur(values, blur)
    if scale == "max":
        values = _process_values_scale_max(values)
    hist.values = values
    if scale == "density":
        hist.normalize()
    return hist


def scale_hist(
    hist: Histogram1D | Histogram, scale: str | float | int = None
) -> Histogram1D | Histogram:
    if scale is None:
        return hist

    if np.max(hist.values) <= 0.0:
        return hist

    if np.sum(hist.values) <= 0.0:
        return hist

    normalization = 1.0
    if scale == "density":
        normalization = np.sum(hist.values) * hist.cell_volume
    elif scale == "max":
        normalization = np.max(hist.values)
    else:
        normalization = scale

    if normalization > 0.0:
        hist.values = hist.values / normalization
    return hist


def plot_1d(
    hist: Histogram1D,
    orientation: str = "vertical",
    kind: str = "line",
    fill: bool = False,
    offset: float = 0.0,
    scale: float = None,
    process_kws: dict = None,
    ax=None,
    **kws,
):
    """Plot one-dimensional histogram.

    Parameters
    ----------
    hist : Histogram1D
        A one-dimensional distribution.
    ax : Axes
        The plotting axis.
    orientation : {"vertical", "horizontal"}
        Whether to plot on the x or y axis.
    kind : {"line", "step", "bar"}
        Whether to plot a line or a piecewise-constant curve.
        "line" calls `ax.plot`, `ax.plotx`, `ax.fill_between`, or `ax.fill_between_x`.
        "step" calls `ax.stairs`.
        "bar" calls `ax.bar` or `ax.barh`.
    fill : bool
        Whether to fill below the curve.
    offset : float
        Offset applied to the profile.
    scale : {None, "density", "max", float}
        Scale the profile by density (area under curve), max value, or value provided.
    process_kws : dict
        Key word arguments passed to `process_hist`.
    **kws
        Key word arguments passed to the plotting function.
    """
    if process_kws is None:
        process_kws = {}

    kws.setdefault("lw", 1.5)

    hist = hist.copy()
    hist = scale_hist(hist, scale=scale)

    values = hist.values.copy()
    coords = hist.coords.copy()
    edges = hist.edges.copy()

    if kind == "step":
        return ax.stairs(
            values + offset,
            edges=edges,
            fill=fill,
            baseline=offset,
            orientation=orientation,
            **kws,
        )
    if kind == "line":
        values = values + offset
        if fill:
            if orientation == "horizontal":
                return ax.fill_betweenx(coords, offset, values, **kws)
            else:
                return ax.fill_between(coords, offset, values, **kws)
        else:
            coords = np.hstack([coords[0], coords, coords[-1]])
            values = np.hstack([offset, values, offset])
            if orientation == "horizontal":
                return ax.plotx(coords, values, **kws)
            else:
                return ax.plot(coords, values, **kws)
    elif kind == "bar":
        pad = offset * np.ones(len(coords))
        if orientation == "horizontal":
            return ax.barh(coords, values, left=pad, **kws)
        else:
            return ax.bar(coords, values, bottom=pad, **kws)
    else:
        raise ValueError(f"Invalid plot kind '{kind}'")

    return ax


def plot_hist_profiles_overlay(
    hist: Histogram,
    profx: bool = True,
    profy: bool = True,
    scale: float = 0.12,
    start: str = "edge",
    keep_limits: bool = False,
    ax=None,
    **kws,
) -> uplt.Axes:
    """Overlay one-dimensional profiles on two-dimensional image.

    Parameters
    ----------
    hist : Histogram
        A two-dimensional histogram.
    profx, profy : bool
        Whether to plot the x/y profile.
    scale : float
        Maximum of the 1D plot relative to the axes limits.
    start : {'edge', 'center'}
        Whether to start the plot at the center or edge of the first row/column.
    **kws
        Key word arguments passed to `psdist.plot.plot_profile`.
    """
    kws.setdefault("kind", "step")
    kws.setdefault("lw", 1.0)
    kws.setdefault("color", "white")  # good for viridis
    kws.setdefault("alpha", 0.6)

    values = hist.values.copy()
    coords = hist.coords

    old_limits = [ax.get_xlim(), ax.get_ylim()]

    for axis, proceed in enumerate([profx, profy]):
        if proceed:
            hist_proj = hist.project(axis)
            hist_proj = scale_hist(hist_proj, scale="max")

            # Scale values based on coordinates on other axis
            other_axis = int(axis == 0)
            hist_proj.values *= scale * hist.ranges[other_axis]

            offset = 0.0
            if start == "edge":
                offset = hist.edges[other_axis][0]
            elif start == "center":
                offset = hist.coords[other_axis][0]

            plot_1d(
                hist=hist_proj,
                ax=ax,
                offset=offset,
                orientation=("horizontal" if axis else "vertical"),
                **kws,
            )
        if keep_limits:
            ax.format(xlim=old_limits[0], ylim=old_limits[1])
    return ax


def plot_rms_ellipse(
    hist: Histogram,
    level: float = 1.0,
    center_at_mean: bool = True,
    ax=None,
    **kws,
) -> uplt.Axes:
    """Compute and plot the RMS ellipse from a two-dimensional image.

    Parameters
    ----------
    hist : Histogram
        A two-dimensional histogram.
    level : number of list of numbers
        If a number, plot the rms ellipse inflated by the number. If a list
        of numbers, repeat for each number.
    center_at_mean : bool
        Whether to center the ellipse at the image centroid.
    **kws
        Key word arguments passed to `psdist.plot.plot_rms_ellipse`.
    """
    cov_matrix = ps.hist.cov(hist)
    mean = (0.0, 0.0)
    if center_at_mean:
        mean = ps.hist.mean(hist)
    return core.plot_rms_ellipse_cov(cov_matrix, mean, level=level, ax=ax, **kws)


def plot(
    hist: Histogram,
    kind: str = "pcolor",
    process_kws: dict = None,
    offset: float = None,
    offset_type: str = "relative",
    mask: bool = False,
    profx: bool = False,
    profy: bool = False,
    prof_kws: dict = None,
    rms_ellipse: bool = False,
    rms_ellipse_kws: dict = None,
    return_mesh: bool = False,
    ax=None,
    **kws,
) -> uplt.Axes:
    """Plot two-dimensional histogram..

    Parameters
    ----------
    hist : Histogram
        A two-dimensional histogram.
    kind : ['pcolor', 'contour', 'contourf']
        Whether to call `ax.pcolormesh`, `ax.contour`, or `ax.contourf`.
    process_kws : dict
        Key word arguments passed to `psdist.image.process`.
    offset: float
        Adds offset to the histogram values (usually to avoid zeros for
         logarithmic colormaps).
    offset_type : {"relative", "absolute"}
        If offset_type is 'relative', adds `min(hist.values) * offset`.
        Otherwise, adds `offset`.
    mask : bool
        Whether to mask zero/negative cells.
    profx, profy : bool
        Whether to plot the x/y profile.
    prof_kws : dict
        Key words arguments for `image_profiles`.
    rms_ellipse : bool
        Whether to plot rms ellipse.
    rms_ellipse_kws : dict
        Key word arguments for `image_rms_ellipse`.
    return_mesh : bool
        Whether to return a mesh from `ax.pcolormesh`.
    ax : Axes
        The axis on which to plot.
    **kws
        Key word arguments passed to plotting function.
    """
    # Process key word arguments.
    if process_kws is None:
        process_kws = {}

    if rms_ellipse_kws is None:
        rms_ellipse_kws = {}
    rms_ellipse_kws.setdefault("color", "white")

    if prof_kws is None:
        prof_kws = {}
    if kind == "contourf":
        prof_kws.setdefault("keep_limits", True)

    kws.setdefault("colorbar", False)
    kws.setdefault("colorbar_kw", {})

    plot_function = None
    if kind == "pcolor":
        plot_function = ax.pcolormesh
        kws.setdefault("ec", "None")
        kws.setdefault("linewidth", 0.0)
        kws.setdefault("rasterized", True)
        kws.setdefault("shading", "auto")
    elif kind == "contour":
        plot_function = ax.contour
    elif kind == "contourf":
        plot_function = ax.contourf
    else:
        raise ValueError("Invalid plot kind.")

    # Process the histogram.
    hist = hist.copy()
    hist = process_hist(hist, **process_kws)

    # Add offset to histogram values.
    if offset is not None:
        if offset_type == "relative":
            if np.count_nonzero(hist.values):
                offset *= np.min(hist.values[hist.values > 0])
        hist.values += offset
    else:
        offset = 0.0

    # Make sure there are no more zero elements if kws["norm"] == "log"
    log = ("norm" in kws) and (kws["norm"] == "log")
    if mask or log:
        hist.values = np.ma.masked_less_equal(hist.values, 0.0)
    if log:
        kws["colorbar_kw"]["formatter"] = "log"

    # If there are only zero elements, increase vmax so that the lowest color shows.
    if not np.count_nonzero(hist.values):
        kws["vmin"] = 1.0
        kws["vmax"] = 1.0

    # Plot the image.
    mesh = plot_function(hist.coords[0], hist.coords[1], hist.values.T, **kws)
    if rms_ellipse:
        plot_rms_ellipse(hist, ax=ax, **rms_ellipse_kws)
    if profx or profy:
        hist.values -= offset
        plot_hist_profiles_overlay(hist, profx=profx, profy=profy, ax=ax, **prof_kws)

    if return_mesh:
        return ax, mesh
    else:
        return ax


def plot_joint(hist: Histogram, grid_kws: dict = None, **kws):
    from psdist.plot.grid import JointGrid

    if grid_kws is None:
        grid_kws = {}

    grid = JointGrid(**grid_kws)
    grid.plot_hist(hist, **kws)
    return grid


def plot_corner(hist: Histogram, grid_kws: dict = None, **kws):
    from psdist.plot.grid import CornerGrid

    if grid_kws is None:
        grid_kws = {}

    ndim = hist.ndim

    grid = CornerGrid(ndim=ndim, **grid_kws)
    grid.plot_hist(hist, **kws)
    return grid


def plot_slice_matrix(hist: Histogram, grid_kws: dict = None, **kws):
    from psdist.plot.grid import SliceGrid

    if grid_kws is None:
        grid_kws = {}
    grid_kws.setdefault("space", 0.2)
    grid_kws.setdefault("annotate_kws_view", dict(color="white"))
    grid_kws.setdefault("annotate_kws_slice", dict(color="black"))
    grid_kws.setdefault("xticks", [])
    grid_kws.setdefault("yticks", [])
    grid_kws.setdefault("xspineloc", "neither")
    grid_kws.setdefault("yspineloc", "neither")

    grid = SliceGrid(**grid_kws)
    grid.plot_hist(hist, **kws)
    return grid


# # [TO DO] Update `proj2d_interactive_slice` and `proj1d_interactive_slice` to include
# # `options` input; see the versions of these functions in `psdist.plot.points`.
#

def plot_interactive_slice_2d(
    hist: Histogram,
    default_ind: tuple[int, int] = (0, 1),
    slice_type: str = "int",
    dims: list[str] = None,
    units: list[str] = None,
    cmaps: list[str] = None,
    thresh_slider: bool = False,
    profiles_checkbox: bool = False,
    fig_kws: dict = None,
    **plot_kws,
):
    """2D partial projection with interactive slicing.

    The distribution is projected onto the specified axes. Sliders provide the
    option to slice the distribution before projecting.

    Parameters
    ----------
    hist : Histogram
        A two-dimensional histogram.
    default_ind : (i, j)
        Default x and y index to plot.
    slice_type : {'int', 'range'}
        Whether to slice one index along the axis or a range of indices.
    dims, units : list[str], shape (n,)
        Dimension names and units.
    fig_kws : dict
        Key words for `uplt.subplots`.
    cmaps : list
        Color map options for dropdown menu.
    thresh_slider : bool
        Whether to include a threshold slider.
    profiles_checkbox : bool
        Whether to include a profiles checkbox.
    **plot_kws
        Key word arguments for `plot.image.plot`.

    Returns
    -------
    ipywidgets.widgets.interaction.interactive
        This widget can be displayed by calling `IPython.display.display(gui)`.
    """
    if fig_kws is None:
        fig_kws = {}

    if dims is None:
        dims = [f"x{i + 1}" for i in range(hist.ndim)]

    if units is None:
        units = hist.ndim * [""]

    dims_units = []
    for dim, unit in zip(dims, units):
        if unit:
            dims_units.append(f"{dim} [{unit}]")
        else:
            dims_units.append(dim)

    plot_kws.setdefault("colorbar", True)
    plot_kws["process_kws"] = dict(thresh_type="frac")

    # Widgets
    cmap = None
    if cmaps is not None:
        cmap = widgets.Dropdown(options=cmaps, description="cmap")

    if thresh_slider:
        thresh_slider = widgets.FloatSlider(
            value=-3.3,
            min=-8.0,
            max=0.0,
            step=0.1,
            description="thresh (log)",
            continuous_update=True,
        )

    discrete = widgets.Checkbox(value=False, description="discrete")

    log = widgets.Checkbox(value=False, description="log")

    _profiles_checkbox = None
    if profiles_checkbox:
        _profiles_checkbox = widgets.Checkbox(value=False, description="profiles")

    dim1 = widgets.Dropdown(options=dims, index=default_ind[0], description="dim 1")
    dim2 = widgets.Dropdown(options=dims, index=default_ind[1], description="dim 2")

    # Sliders and checkboxes (for slicing). Each unplotted dimension has a
    # checkbox which determine if that dimension is sliced. The slice
    # indices are determined by the slider.
    sliders, checks = [], []
    for k in range(hist.ndim):
        if slice_type == "int":
            slider = widgets.IntSlider(
                min=0,
                max=hist.shape[k],
                value=(hist.shape[k] // 2),
                description=dims[k],
                continuous_update=True,
            )
        elif slice_type == "range":
            slider = widgets.IntRangeSlider(
                value=(0, hist.shape[k]),
                min=0,
                max=hist.shape[k],
                description=dims[k],
                continuous_update=True,
            )
        else:
            raise ValueError("`slice_type` must be 'int' or 'range'.")

        slider.layout.display = "none"
        sliders.append(slider)
        checks.append(widgets.Checkbox(description=f"slice {dims[k]}"))

    def hide(button):
        """Hide/show sliders."""
        for k in range(hist.ndim):
            # Hide elements for dimensions being plotted.
            valid = dims[k] not in (dim1.value, dim2.value)
            disp = None if valid else "none"
            for element in [sliders[k], checks[k]]:
                element.layout.display = disp

            # Uncheck boxes for dimensions being plotted.
            if not valid and checks[k].value:
                checks[k].value = False

            # Make sliders respond to check boxes.
            if not checks[k].value:
                sliders[k].layout.display = "none"

    # Update the slider list automatically.
    for element in (dim1, dim2, *checks):
        element.observe(hide, names="value")

    # Initial hide
    for k in range(hist.ndim):
        if k in default_ind:
            checks[k].layout.display = "none"
            sliders[k].layout.display = "none"

    def update(**kws):
        """Update the figure."""
        dim1 = kws["dim1"]
        dim2 = kws["dim2"]

        ind, checks = [], []
        for i in range(100):
            if f"check{i}" in kws:
                checks.append(kws[f"check{i}"])
            if f"slider{i}" in kws:
                ind.append(kws[f"slider{i}"])

        # Return nothing if input does not make sense.
        for dim, check in zip(dims, checks):
            if check and dim in (dim1, dim2):
                return
        if dim1 == dim2:
            return

        # Slice and project the distribution.
        axis_view = [dims.index(dim) for dim in (dim1, dim2)]
        axis_slice = [dims.index(dim) for dim, check in zip(dims, checks) if check]
        for k in range(hist.ndim):
            if type(ind[k]) is int:
                ind[k] = (ind[k], ind[k] + 1)
        ind = [ind[k] for k in axis_slice]

        hist_proj = hist.slice(axis=axis_slice, ind=ind).project(axis_view)

        # Update plotting key word arguments.
        if "cmap" in kws:
            plot_kws["cmap"] = kws["cmap"]
        if "profiles_checkbox" in kws:
            plot_kws["profx"] = plot_kws["profy"] = kws["profiles_checkbox"]
        plot_kws["norm"] = "log" if kws["log"] else None
        plot_kws["process_kws"]["fill_value"] = 0
        if "thresh_slider" in kws:
            plot_kws["process_kws"]["thresh"] = 10.0 ** kws["thresh_slider"]
        else:
            plot_kws["process_kws"]["thresh"] = None

        # Plot the projection onto the specified axes.
        fig, ax = uplt.subplots(**fig_kws)
        plot(hist_proj, ax=ax, **plot_kws)
        ax.format(xlabel=dims_units[axis_view[0]], ylabel=dims_units[axis_view[1]])
        plt.show()

    # Pass key word arguments for `update`.
    kws = {}
    if cmap is not None:
        kws["cmap"] = cmap
    if profiles_checkbox:
        kws["profiles_checkbox"] = _profiles_checkbox
    kws["log"] = log
    kws["dim1"] = dim1
    kws["dim2"] = dim2
    if thresh_slider:
        kws["thresh_slider"] = thresh_slider
    for i, check in enumerate(checks, start=1):
        kws[f"check{i}"] = check
    for i, slider in enumerate(sliders, start=1):
        kws[f"slider{i}"] = slider
    return interactive(update, **kws)


def plot_interactive_slice_1d(
    hist: Histogram,
    default_ind: int = 0,
    slice_type: str = "int",
    dims: list[str] = None,
    units: list[str] = None,
    fig_kws: dict = None,
    **plot_kws,
):
    """1D partial projection with interactive slicing.

    Parameters
    ----------
    hist : Histogram
        A two-dimensional histogram.
    default_ind : int
        Default index to plot.
    slice_type : {'int', 'range'}
        Whether to slice one index along the axis or a range of indices.
    dims, units : list[str], shape (n,)
        Dimension names and units.
    kind : {'bar', 'line'}
        The kind of plot to draw.
    fig_kws : dict
        Key word arguments passed to `proplot.subplots`.
    **plot_kws
        Key word arguments passed to 1D plotting function.

    Returns
    -------
    gui : ipywidgets.widgets.interaction.interactive
        This widget can be displayed by calling `IPython.display.display(gui)`.
    """
    if dims is None:
        dims = [f"x{i + 1}" for i in range(hist.ndim)]

    if units is None:
        units = hist.ndim * [""]

    dims_units = []
    for dim, unit in zip(dims, units):
        if unit:
            dims_units.append(f"{dim} [{unit}]")
        else:
            dims_units.append(dim)

    if fig_kws is None:
        fig_kws = {}
    fig_kws.setdefault("figsize", (4.5, 1.5))

    plot_kws.setdefault("color", "black")
    plot_kws.setdefault("kind", "stepfilled")

    # Widgets
    dim1 = widgets.Dropdown(options=dims, index=default_ind, description="dim")

    # Sliders
    sliders, checks = [], []
    for k in range(hist.ndim):
        if slice_type == "int":
            slider = widgets.IntSlider(
                min=0,
                max=hist.shape[k],
                value=hist.shape[k] // 2,
                description=dims[k],
                continuous_update=True,
            )
        elif slice_type == "range":
            slider = widgets.IntRangeSlider(
                value=(0, hist.shape[k]),
                min=0,
                max=hist.shape[k],
                description=dims[k],
                continuous_update=True,
            )
        else:
            raise ValueError("Invalid `slice_type`.")
        slider.layout.display = "none"
        sliders.append(slider)
        checks.append(widgets.Checkbox(description=f"slice {dims[k]}"))

    def hide(button):
        """Hide/show sliders based on checkboxes."""
        for k in range(hist.ndim):
            # Hide elements for dimensions being plotted.
            valid = dims[k] != dim1.value
            disp = None if valid else "none"
            for element in [sliders[k], checks[k]]:
                element.layout.display = disp

            # Uncheck boxes for dimensions being plotted.
            if not valid and checks[k].value:
                checks[k].value = False

            # Make sliders respond to check boxes.
            if not checks[k].value:
                sliders[k].layout.display = "none"

    # Update the slider list automatically.
    for element in (dim1, *checks):
        element.observe(hide, names="value")

    # Initial hide
    for k in range(hist.ndim):
        if k == default_ind:
            checks[k].layout.display = "none"
            sliders[k].layout.display = "none"

    def update(**kws):
        """Update the figure."""
        dim1 = kws["dim1"]

        ind, checks = [], []
        for i in range(100):
            if f"check{i}" in kws:
                checks.append(kws[f"check{i}"])
            if f"slider{i}" in kws:
                ind.append(kws[f"slider{i}"])

        # Return nothing if input does not make sense.
        for dim, check in zip(dims, checks):
            if check and dim == dim1:
                return

        # Slice, then project onto the specified axis.
        axis_view = dims.index(dim1)
        axis_slice = [dims.index(dim) for dim, check in zip(dims, checks) if check]
        for k in range(hist.ndim):
            if type(ind[k]) is int:
                ind[k] = (ind[k], ind[k] + 1)
        ind = [ind[k] for k in axis_slice]

        hist_proj = hist.slice(axis=axis_slice, ind=ind).project(axis_view)
        hist_proj.normalize()

        # Plot the projection.
        fig, ax = uplt.subplots(**fig_kws)
        plot_1d(hist_proj, ax=ax, **plot_kws)
        ax.format(xlabel=dims_units[axis_view])
        plt.show()

    kws = {"dim1": dim1}
    for i, check in enumerate(checks, start=1):
        kws[f"check{i}"] = check
    for i, slider in enumerate(sliders, start=1):
        kws[f"slider{i}"] = slider
    return interactive(update, **kws)



"""Covariance matrix analysis."""

import numpy as np


def rms_ellipsoid_volume(S: np.ndarray) -> float:
    return np.sqrt(np.linalg.det(S))


def projected_emittances(S: np.ndarray) -> tuple[float, ...]:
    emittances = []
    for i in range(0, S.shape[0], 2):
        emittance = rms_ellipsoid_volume(S[i : i + 2, i : i + 2])
        emittances.append(emittance)
    return emittances


def intrinsic_emittances(S: np.ndarray) -> tuple[float, ...]:
    S = S[:4, :4].copy()  # [to do]: expand to NxN
    U = np.array(
        [
            [+0.0, +1.0, +0.0, +0.0],
            [-1.0, +0.0, +0.0, +0.0],
            [+0.0, +0.0, +0.0, +1.0],
            [+0.0, +0.0, -1.0, +0.0],
        ]
    )
    tr_SU2 = np.trace(np.linalg.matrix_power(np.matmul(S, U), 2))
    det_S = np.linalg.det(S)
    eps_1 = 0.5 * np.sqrt(-tr_SU2 + np.sqrt(tr_SU2**2 - 16.0 * det_S))
    eps_2 = 0.5 * np.sqrt(-tr_SU2 - np.sqrt(tr_SU2**2 - 16.0 * det_S))
    return (eps_1, eps_2)


def twiss_2d(S: np.ndarray) -> tuple[float, float, float]:
    """Compute twiss parameters from 2 x 2 covariance matrix.

    alpha = -<xx'> / sqrt(<xx><x'x'> - <xx'><xx'>)
    beta  =  <xx>  / sqrt(<xx><x'x'> - <xx'><xx'>)
    """
    emittance = rms_ellipsoid_volume(S)
    beta = S[0, 0] / emittance
    alpha = -S[0, 1] / emittance
    return (alpha, beta, emittance)


def twiss(
    S: np.ndarray,
) -> tuple[float, float, float] | list[tuple[float, float, float]]:
    parameters = []
    for i in range(0, S.shape[0], 2):
        parameters.append(twiss_2d(S[i : i + 2, i : i + 2]))

    if len(parameters) == 1:
        parameters = parameters[0]

    return parameters


def unit_symplectic_matrix(ndim: int) -> np.ndarray:
    """Return matrix U such that, if M is a symplectic matrix, UMU^T = M."""
    U = np.zeros((ndim, ndim))
    for i in range(0, ndim, 2):
        U[i : i + 2, i : i + 2] = [[0.0, 1.0], [-1.0, 0.0]]
    return U


def normalize_eigvecs(eigvecs: np.ndarray) -> np.ndarray:
    """Normalize eigenvectors according to Lebedev-Bogacz convention."""
    ndim = eigvecs.shape[0]
    U = unit_symplectic_matrix(ndim)
    for i in range(0, ndim, 2):
        v = eigvecs[:, i]
        val = np.linalg.multi_dot([np.conj(v), U, v]).imag
        if val > 0.0:
            (eigvecs[:, i], eigvecs[:, i + 1]) = (eigvecs[:, i + 1], eigvecs[:, i])
        eigvecs[:, i : i + 2] *= np.sqrt(2.0 / np.abs(val))
    return eigvecs


def normalization_matrix_from_eigvecs(eigvecs: np.ndarray) -> np.ndarray:
    """Return normalization matrix V^-1 from eigenvectors."""
    V = np.zeros(eigvecs.shape)
    for i in range(0, V.shape[1], 2):
        V[:, i] = eigvecs[:, i].real
        V[:, i + 1] = (1.0j * eigvecs[:, i]).real
    return np.linalg.inv(V)


def normalization_matrix_from_twiss_2d(
    alpha: float, beta: float, emittance: float = None
) -> np.ndarray:
    """Return 2 x 2 normalization matrix V^-1 from Twiss parameters."""
    V = np.array([[beta, 0.0], [-alpha, 1.0]]) * np.sqrt(1.0 / beta)
    A = np.eye(2)
    if emittance is not None:
        A = np.sqrt(np.diag([emittance, emittance]))
    V = np.matmul(V, A)
    return np.linalg.inv(V)


def normalization_matrix_from_twiss(
    twiss_params: list[tuple[float, float, float]]
) -> np.ndarray:
    """2N x 2N block-diagonal normalization matrix from Twiss parameters.

    Parameters
    ----------
    twiss_params : list[tuple[float, float, float]]
        Twiss parameters (alpha, beta, emittance) in each dimension.

    Returns
    -------
    V : ndarray, shape (2N, 2N)
        Block-diagonal normalization matrix.
    """
    ndim = len(twiss_params) // 2
    V = np.zeros((ndim, ndim))
    for i in range(0, ndim, 2):
        V[i : i + 2, i : i + 2] = normalization_matrix_from_twiss_2d(
            *twiss_params[i : i + 2]
        )
    return np.linalg.inv(V)


def normalization_matrix(
    S: np.ndarray, scale: bool = False, block_diag: bool = False
) -> np.ndarray:
    """Return normalization matrix V^{-1} from covariance matrix S.

    Parameters
    ----------
    S : np.ndarray
        An N x N covariance matrix.
    scale : bool
        If True, normalize to unit rms emittance.
    block_diag : bool
        If true, normalize only 2x2 block-diagonal elements (x-x', y-y', etc.).
    """

    def _normalization_matrix(S: np.ndarray, scale: bool = False) -> np.ndarray:
        U = unit_symplectic_matrix(ndim)
        SU = np.matmul(S, U)
        eigvals, eigvecs = np.linalg.eig(SU)
        eigvecs = normalize_eigvecs(eigvecs)
        V_inv = normalization_matrix_from_eigvecs(eigvecs)

        if scale:
            V = np.linalg.inv(V_inv)
            A = np.eye(ndim)
            for i in range(0, ndim, 2):
                emittance = rms_ellipsoid_volume(S[i : i + 2, i : i + 2])
                A[i : i + 2, i : i + 2] *= np.sqrt(emittance)
            V = np.matmul(V, A)
            V_inv = np.linalg.inv(V)

        return V_inv

    ndim = S.shape[0]
    V_inv = np.eye(ndim)
    if block_diag:
        V_inv = _normalization_matrix(S, scale=scale)
    else:
        for i in range(0, ndim, 2):
            V_inv[i : i + 2, i : i + 2] = _normalization_matrix(
                S[i : i + 2, i : i + 2], scale=scale
            )
    return V_inv


def cov_to_corr(S: np.ndarray) -> np.ndarray:
    """Compute correlation matrix from covariance matrix."""
    D = np.sqrt(np.diag(S.diagonal()))
    Dinv = np.linalg.inv(D)
    return np.linalg.multi_dot([Dinv, S, Dinv])


def rms_ellipse_params(
    S: np.ndarray, axis: tuple[int, ...] = None
) -> tuple[float, ...]:
    """Return projected rms ellipse dimensions and orientation.

    Parameters
    ----------
    S : ndarray, shape (2N, 2N)
        The phase space covariance matrix.
    axis : tuple[int]
        Projection axis. Example: if the axes are {x, xp, y, yp}, and axis=(0, 2),
        the four-dimensional ellipsoid is projected onto the x-y plane.

    Returns
    -------
    c1, c2 : float
        The ellipse semi-axis widths.
    angle : float
        The tilt angle below the x axis [radians].
    """
    if S.shape[0] == 2:
        axis = (0, 1)
    (i, j) = axis
    sii = S[i, i]
    sjj = S[j, j]
    sij = S[i, j]
    angle = -0.5 * np.arctan2(2 * sij, sii - sjj)
    _sin = np.sin(angle)
    _cos = np.cos(angle)
    _sin2 = _sin**2
    _cos2 = _cos**2
    c1 = np.sqrt(abs(sii * _cos2 + sjj * _sin2 - 2 * sij * _sin * _cos))
    c2 = np.sqrt(abs(sii * _sin2 + sjj * _cos2 + 2 * sij * _sin * _cos))
    return (c1, c2, angle)



import warnings
from typing import Callable
from typing import Union

import numpy as np
import matplotlib.pyplot as plt
import ultraplot as uplt

from ..hist import Histogram
from ..hist import Histogram1D
from ..core import histogram as _histogram
from ..core import limits as _get_limits
from ..hist import Histogram
from ..utils import array_like
from .hist import plot as _plot_hist, scale_hist
from .hist import plot_1d as _plot_hist_1d
from .points import plot as _plot_points


class JointGrid:
    """Grid for joint plots.

    Attributes
    -----------
    fig : proplot.figure.Figure
        The main figure.
    ax : proplot.gridspec.SubplotGrid
        The main axis.
    ax_panel_x, ax_panel_y : proplot.gridspec.SubplotGrid
        The marginal (panel) axes on the top and right.
    """

    def __init__(
        self,
        limits: np.ndarray = None,
        panel_kws: dict = None,
        panel_fmt_kws: dict = None,
        panel_fmt_kws_x: dict = None,
        panel_fmt_kws_y: dict = None,
        **fig_kws,
    ) -> None:
        """Constructor.

        panel_kws : dict
            Key word arguments for `ax.panel`.
        panel_fmt_kws, panel_fmt_kws_x, panel_fmt_kws_y : dict
            Key word arguments for `ax.format` for each of the marginal (panel) axs.
        **fig_kws
            Key word arguments passed to `proplot.subplots`.
        """
        self.fig, self.ax = uplt.subplots(**fig_kws)

        self.limits = limits
        if self.limits is not None:
            self.set_limits(self.limits)

        if panel_kws is None:
            panel_kws = {}
        if panel_fmt_kws is None:
            panel_fmt_kws = {}
        if panel_fmt_kws_x is None:
            panel_fmt_kws_x = {}
        if panel_fmt_kws_y is None:
            panel_fmt_kws_y = {}

        panel_fmt_kws_x.setdefault("xspineloc", "bottom")
        panel_fmt_kws_x.setdefault("yspineloc", "left")
        panel_fmt_kws_y.setdefault("xspineloc", "bottom")
        panel_fmt_kws_y.setdefault("yspineloc", "left")

        self.ax_panel_x = self.ax.panel("t", **panel_kws)
        self.ax_panel_y = self.ax.panel("r", **panel_kws)
        self.panel_axs = [self.ax_panel_x, self.ax_panel_y]
        for ax in self.panel_axs:
            ax.format(**panel_fmt_kws)
        self.panel_axs[0].format(**panel_fmt_kws_x)
        self.panel_axs[1].format(**panel_fmt_kws_y)

        self.default_panel_plot_kws = {
            "color": "black",
            "kind": "step",
            "lw": 1.5,
            "scale": "density",
        }

        self.default_panel_hist_kws = {"bins": 64}

        self.frozen_limits = False

    def set_limits(self, limits: np.ndarray) -> None:
        self.limits = limits
        self.axs.format(xlim=limits[0], ylim=limits[1])

    def reset_limits(self) -> None:
        self.set_limits(self.limits)

    def freeze_limits(self) -> None:
        self.frozen_limits = True

    def unfreeze_limits(self) -> None:
        self.frozen_limits = False

    def wrapup(self) -> None:
        if self.frozen_limits:
            self.reset_limits()

    def plot(self, points: np.ndarray, **kws):
        return self.plot_points(points, **kws)

    def plot_points(
        self,
        points: np.ndarray,
        panel_hist_kws: dict = None,
        panel_plot_kws: dict = None,
        **kws,
    ) -> None:
        """Plot 2D points.

        Parameters
        ----------
        points: np.ndarray, shape (..., 2)
            Particle coordinates
        panel_hist_kws : dict
            Key word arguments passed to `psdist.points.histogram`.
        panel_plot_kws : dict
            Key word arguments passed to `plot_hist_1d`.
        **kws
            Key word arguments passed to `plot_points.`
        """
        if panel_plot_kws is None:
            panel_plot_kws = {}

        if panel_hist_kws is None:
            panel_hist_kws = {}

        for key, val in self.default_panel_hist_kws.items():
            panel_hist_kws.setdefault(key, val)

        for key, val in self.default_panel_plot_kws.items():
            panel_plot_kws.setdefault(key, val)

        # Plot marginal distributions
        for axis in range(2):
            hist_proj = _histogram(points[:, axis], **panel_hist_kws)
            bins = hist_proj.size

            _plot_hist_1d(
                hist_proj,
                ax=self.panel_axs[axis],
                orientation=("horizontal" if bool(axis) else "vertical"),
                **panel_plot_kws,
            )

        # Plot joint distribution
        kws.setdefault("kind", "hist")
        if kws["kind"] != "scatter":
            if "bins" in kws:
                panel_hist_kws.setdefault("bins", kws["bins"])
            kws.setdefault("bins", bins)
            kws.setdefault("colorbar_kw", {})
            kws["colorbar_kw"].setdefault("pad", 2.0)

        _plot_points(points, ax=self.ax, **kws)

        self.wrapup()

    def plot_hist(
        self,
        hist: Histogram,
        panel_plot_kws: dict = None,
        **kws,
    ) -> None:
        """Plot a two-dimensional histogram.

        Parameters
        ----------
        hist: Histogram
            A two-dimensional histogram.
        panel_plot_kws : dict
            Key word arguments passed to `plot_hist_1d`.
        **kws
            Key word arguments passed to `plot_hist.`
        """
        if panel_plot_kws is None:
            panel_plot_kws = {}

        for key, val in self.default_panel_plot_kws.items():
            panel_plot_kws.setdefault(key, val)

        kws.setdefault("kind", "pcolor")
        kws.setdefault("colorbar_kw", {})
        kws["colorbar_kw"].setdefault("pad", 2.0)

        # Plot marginal distributions
        for axis in range(2):
            hist_proj = hist.project(axis)
            _plot_hist_1d(
                hist_proj,
                ax=self.panel_axs[axis],
                orientation=("horizontal" if bool(axis) else "vertical"),
                **panel_plot_kws,
            )

        # Plot joint distribution
        _plot_hist(hist, ax=self.ax, **kws)

        self.wrapup()

    def colorbar(self, mappable, **kws):
        """Add a colorbar."""
        kws.setdefault("loc", "r")
        kws.setdefault("pad", 2.0)
        self.fig.colorbar(mappable, **kws)


class CornerGrid:
    """Grid for corner plots.

    Attributes
    ----------
    fig : proplot.figure.Figure
        The main figure.
    axs : proplot.gridspec.SubplotGrid
        The subplot axes.
    diag_axs : list[proplot.gridspec.SubplotGrid]
        The axes for diagonal (univariate) subplots. Can be empty.
    offdiag_axs : list[proplot.gridspec.SubplotGrid]
        The axes for off-diagonal (bivariate) subplots.
    diag_indices : list[int]
        The index of the dimension plotted on each diagonal subplot.
    offdiag_indices : list[2-tuple of int]
        Indices of the dimensions plotted on each off-diagonal subplot.
    """

    def __init__(
        self,
        ndim: int,
        diag: bool = True,
        diag_shrink: float = 1.0,
        diag_rspine: bool = False,
        diag_share: bool = False,
        diag_frozen: bool = False,
        limits: list[tuple[float, float]] = None,
        labels: list[str] = None,
        corner: bool = True,
        **fig_kws,
    ) -> None:
        """
        Parameters
        ----------
        ndim : int
            The number of rows/columns.
        diag : bool
            Whether to include diagonal subplots (univariate plots). If False,
            we have an (N - 1) x (N - 1) grid instead of an N x N grid.
        diag_shrink : float in range [0, 1]
            Scales the maximum y value of the diagonal profile plots.
        diag_rspine : bool
            Whether to include right spine on diagonal subplots (if `corner`).
        diag_share : bool
            Whether to share diagonal axis limits; i.e., whether we can compare
            the areas under the diagonal profile plots.
        diag_frozen : bool
            If true, don't adjust the y limits on diagonal axes after the first
            distribution is plotted. Otherwise, always adjust to keep the
            histograms in view.
        limits : list[tuple], length n
            The (min, max) for each dimension. (These can be set later.)
        labels : list[str]
            The label for each dimension. (These can be set later.)
        corner : bool
            Whether to hide the upper-triangular subplots.
        **fig_kws
            Key word arguments passed to `uplt.subplots()`.
        """
        # Create figure.
        self.new = True
        self.ndim = self.nrows = self.ncols = ndim
        self.corner = corner

        self.diag = diag
        self.diag_shrink = diag_shrink
        self.diag_rspine = diag_rspine
        self.diag_share = diag_share
        self.diag_ymin = None
        self.diag_yscale = self.ndim * [None]
        self.diag_frozen = diag_frozen

        if not self.diag:
            self.nrows = self.nrows - 1
            self.ncols = self.ncols - 1

        self.fig_kws = fig_kws
        self.fig_kws.setdefault("figwidth", 1.5 * self.nrows)
        self.fig_kws.setdefault("aligny", True)

        self.fig, self.axs = uplt.subplots(
            nrows=self.nrows,
            ncols=self.ncols,
            spanx=False,
            spany=False,
            sharex=False,
            sharey=False,
            **self.fig_kws,
        )

        # Collect diagonal/off-diagonal subplots and indices.
        self.diag_axs = []
        self.offdiag_axs = []
        self.offdiag_axs_u = []

        self.diag_indices = []
        self.offdiag_indices = []
        self.offdiag_indices_u = []

        if self.diag:
            for i in range(self.ndim):
                self.diag_axs.append(self.axs[i, i])
                self.diag_indices.append(i)
            for i in range(1, self.ndim):
                for j in range(i):
                    self.offdiag_axs.append(self.axs[i, j])
                    self.offdiag_axs_u.append(self.axs[j, i])
                    self.offdiag_indices.append((j, i))
                    self.offdiag_indices_u.append((i, j))
        else:
            for i in range(self.ndim - 1):
                for j in range(i + 1):
                    self.offdiag_axs.append(self.axs[i, j])
                    self.offdiag_indices.append((j, i + 1))

        # Set limits and labels.
        self.limits = limits
        if limits is not None:
            self.set_limits(limits)

        self.labels = labels
        if labels is not None:
            self.set_labels(labels)

        # Formatting
        if self.corner or not self.diag:
            for i in range(self.nrows):
                for j in range(self.ncols):
                    if j > i:
                        self.axs[i, j].axis("off")

        self.axs[:-1, :].format(xticklabels=[])
        for i in range(self.nrows):
            for j in range(self.ncols):
                ax = self.axs[i, j]
                if i != self.nrows - 1:
                    ax.format(xticklabels=[])
                if j != 0:
                    if not (i == j and self.diag_rspine and self.corner and self.diag):
                        ax.format(yticklabels=[])

        self.axs.format(xspineloc="bottom", yspineloc="left")
        if self.corner:
            if self.diag_rspine:
                self.format_diag(yspineloc="right")
            else:
                self.format_diag(yspineloc="neither")
        self.axs.format(
            xtickminor=True, ytickminor=True, xlocator=("maxn", 3), ylocator=("maxn", 3)
        )
        self.set_diag_scale("linear")

    def format_diag(self, **kws) -> None:
        """Format diagonal subplots."""
        for ax in self.diag_axs:
            ax.format(**kws)
        if not self.corner:
            for ax in self.diag_axs[1:]:
                ax.format(yticklabels=[])
        self._fake_diag_yticks()

    def format_offdiag(self, **kws) -> None:
        """Format off-diagonal subplots."""
        for ax in [self.offdiag_axs + self.offdiag_axs_u]:
            ax.format(**kws)

    def get_labels(self) -> list[str]:
        """Return the dimension labels."""
        if self.diag:
            labels = [ax.get_xlabel() for ax in self.diag_axs]
        else:
            labels = [self.axs[-1, i].get_xlabel() for i in range(self.ndim - 1)]
            labels = labels + [self.axs[-1, 0].get_ylabel()]
        return labels

    def set_labels(self, labels: list[str]) -> None:
        """Set the dimension labels."""
        for ax, label in zip(self.axs[-1, :], labels):
            ax.format(xlabel=label)
        for ax, label in zip(self.axs[int(self.diag) :, 0], labels[1:]):
            ax.format(ylabel=label)
        if self.diag and not self.corner:
            self.axs[0, 0].format(ylabel=labels[0])
        self.labels = labels

    def get_limits(self) -> list[tuple[float, float]]:
        """Return the plot limits."""
        if self.diag:
            limits = [ax.get_xlim() for ax in self.diag_axs]
        else:
            limits = [self.axs[-1, i].get_xlim() for i in range(self.ndim - 1)]
            limits = limits + [self.axs[-1, 0].get_ylim()]
        return limits

    def set_limits(
        self, limits: list[tuple[float, float]], expand: bool = False
    ) -> None:
        """Set the plot limits.

        Parameters
        ----------
        limits : list[tuple], length n
            The (min, max) for each dimension.
        expand : bool
            If True, compare the proposed limits to the existing limits, expanding
            if the new limits are wider.
        """
        if limits is not None:
            if expand:
                limits = np.array(limits)
                limits_old = np.array(self.get_limits())
                mins = np.minimum(limits[:, 0], limits_old[:, 0])
                maxs = np.maximum(limits[:, 1], limits_old[:, 1])
                limits = list(zip(mins, maxs))

            if self.diag:
                for i in range(self.ndim):
                    for j in range(self.ndim):
                        if i != j:
                            if (j < i) or (not self.corner):
                                self.axs[i, j].format(ylim=limits[i])
            else:
                for i in range(self.ndim - 1):
                    for ax in self.axs[i, :]:
                        ax.format(ylim=limits[i + 1])

            for i in range(self.ncols):
                self.axs[:, i].format(xlim=limits[i])

        self.limits = self.get_limits()

    def get_default_diag_kws(self, diag_kws: dict = None) -> dict:
        if diag_kws is None:
            diag_kws = {}
        diag_kws.setdefault("color", "black")
        diag_kws.setdefault("lw", 1.0)
        diag_kws.setdefault("kind", "step")
        diag_kws.setdefault("scale", "density")
        return diag_kws

    def _fake_diag_yticks(self) -> None:
        """The yticks on the (0, 0) subplot correspond to the other subplots in the row.

        Source: pandas.plotting.scatterplot_matrix.
        """
        if self.corner or not self.diag:
            return
        limits = self.limits
        if limits is None:
            limits = self.get_limits()

        lim1 = limits[0]
        locs = self.axs[0, 0].xaxis.get_majorticklocs()
        locs = locs[(lim1[0] <= locs) & (locs <= lim1[1])]
        adj = (locs - lim1[0]) / (lim1[1] - lim1[0])

        lim0 = self.axs[0, 0].get_ylim()
        adj = adj * (lim0[1] - lim0[0]) + lim0[0]
        self.axs[0, 0].yaxis.set_ticks(adj)

        if np.all(locs == locs.astype(int)):
            locs = locs.astype(int)
        self.axs[0, 0].yaxis.set_ticklabels(locs)

    def _force_non_negative_diag_ymin(self) -> None:
        """Force diagonal ymins to be at least zero."""
        if any([ax.get_ylim()[0] < 0.0 for ax in self.diag_axs]):
            self.format_diag(ylim=0.0)

    def set_diag_scale(self, scale: str = "linear", pad: float = 0.05) -> None:
        """Set diagonal axis scale.

        Parameters
        ----------
        scale : {"linear", "log"}
            If "linear", scale runs from 0 to 1. If "log", scale runs from half the
            minimum plotted value to 1.
        pad: float
            Padding applied to the y axis limit.
        """
        if scale == "linear":
            ymin = 0.0
            ymax = 1.0 / self.diag_shrink
            delta = ymax - ymin
            ymax = ymin + delta * (1.0 + pad)
            self.format_diag(yscale="linear", yformatter="auto", ymin=ymin, ymax=ymax)
        elif scale == "log":
            ymin = self.diag_ymin
            ymax = 1.0
            log_ymin = np.log10(ymin)
            log_ymax = np.log10(ymax)
            log_delta = log_ymax - log_ymin
            log_delta = log_delta / self.diag_shrink
            ymax = 10.0 ** (log_ymin + log_delta * (1.0 + pad))
            self.format_diag(yscale="log", yformatter="log", ymin=ymin, ymax=ymax)

    def plot_diag(self, hists: list[Histogram1D], **kws) -> None:
        """Compute one-dimensional histograms."""
        if "scale" in kws:
            kws.pop("scale")

        # Normalize histograms
        for hist in hists:
            hist.normalize()

        # Update diag_yscale
        if not (self.diag_frozen and not self.new):
            for axis in range(self.ndim):
                if self.diag_yscale[axis] is None:
                    self.diag_yscale[axis] = -np.inf
            if self.diag_share:
                max_value = np.max([np.max(hist.values) for hist in hists])
                self.diag_yscale = np.maximum(self.diag_yscale, max_value)
            else:
                max_values = np.array([np.max(hist.values) for hist in hists])
                self.diag_yscale = np.maximum(self.diag_yscale, max_values)

        # Plot histograms
        for axis, hist in enumerate(hists):

            yscale = self.diag_yscale[axis]
            if yscale is None:
                yscale = 1.0

            hist_scaled = hist.copy()
            hist_scaled.values /= yscale

            _plot_hist_1d(hist_scaled, ax=self.diag_axs[axis], **kws)

        # Compute minimum positive value (for log scaling)
        for axis, hist in enumerate(hists):
            if not self.diag_ymin:
                self.diag_ymin = np.inf
            self.diag_ymin = min(self.diag_ymin, np.min(hist.values[hist.values > 0.0]))

    def plot_hist(
        self,
        hist: Histogram,
        prof_edge_only: bool = False,
        lower: bool = True,
        upper: bool = True,
        diag: bool = True,
        update_limits: bool = True,
        diag_kws: dict = None,
        **kws,
    ) -> None:
        """Plot an image.

        Parameters
        ----------
        hist: np.ndarray
            An N-dimensional histogram.
        prof_edge_only : bool
            If plotting profiles on top of images (on off-diagonal subplots), whether
            to plot x profiles only in bottom row and y profiles only in left column.
        lower, upper, diag : bool
            Whether to plot on the lower triangular, upper triangular, and/or diagonal subplots.
        update_limits : bool
            Whether to extend the existing plot limits.
        diag_kws : dict
            Key word argument passed to `plot_hist_1d`.
        **kws
            Key word arguments pass to `plot_hist`
        """
        kws.setdefault("kind", "pcolor")

        diag_kws = self.get_default_diag_kws(diag_kws)

        if update_limits:
            self.set_limits(hist.limits, expand=(not self.new))

        # Univariate plots
        if self.diag and diag:
            hists = [hist.project(axis) for axis in range(hist.ndim)]
            self.plot_diag(hists, **diag_kws)

        # Bivariate plots
        profx = kws.get("profx", False)
        profy = kws.pop("profy", False)

        if lower:
            for ax, axis in zip(self.offdiag_axs, self.offdiag_indices):
                hist_proj = hist.project(axis)

                if prof_edge_only:
                    if profx:
                        kws["profx"] = axis[1] == self.ndim - 1
                    if profy:
                        kws["profy"] = axis[0] == 0

                _plot_hist(hist_proj, ax=ax, **kws)

        if upper and not self.corner:
            for ax, axis in zip(self.offdiag_axs_u, self.offdiag_indices_u):
                _plot_hist(hist.project(axis), ax=ax, **kws)

        self._post_plot()

    def plot(self, points: np.ndarray, **kws) -> None:
        return self.plot_points(points, **kws)

    def plot_points(
        self,
        points: np.ndarray,
        limits: list[tuple[float, float]] = None,
        bins: int = 50,
        prof_edge_only: bool = False,
        lower: bool = True,
        upper: bool = True,
        diag: bool = True,
        autolim_kws: dict = None,
        update_limits: bool = True,
        diag_kws: dict = None,
        **kws,
    ) -> None:
        """Plot points.

        Parameters
        ----------
        points : np.ndarray, shape (..., N)
            Particle coordinates.
        limits : list[tuple[float, float]], length N
            The (min, max) axis limits.
        bins : 'auto', int, list[int]
            The number of bins along each dimension (if plot type requires histogram
            computation). If int or 'auto', applies to all dimensions.
        prof_edge_only : bool
            If plotting profiles on top of images (on off-diagonal subplots), whether
            to plot x profiles only in bottom row and y profiles only in left column.
        lower, upper, diag : bool
            Whether to plot on the lower triangular, upper triangular, and/or diagonal subplots.
        update_limits : bool
            Whether to extend the existing plot limits.
        diag_kws : dict
            Key word argument passed to `plot.plot_profile`.
        **kws
            Key word arguments pass to `plot.points.plot`
        """
        kws.setdefault("kind", "hist")

        diag_kws = self.get_default_diag_kws(diag_kws)

        if autolim_kws is None:
            autolim_kws = {}
        autolim_kws.setdefault("pad", 0.25)

        if limits is None:
            limits = _get_limits(points, **autolim_kws)

        if update_limits:
            self.set_limits(limits, expand=(not self.new))

        limits = self.get_limits()

        if not array_like(bins):
            bins = points.shape[1] * [bins]

        # Compute histogram bin edges.
        hists = []
        for axis in range(self.ndim):
            if array_like(bins[axis]):
                edges = bins[axis]
            else:
                edges = np.histogram_bin_edges(
                    points[:, axis], bins[axis], limits[axis]
                )
            hist = Histogram1D(edges=edges)
            hist.bin(points[:, axis])
            hists.append(hist)

        # Univariate plots.
        if self.diag and diag:
            self.plot_diag(hists, **diag_kws)

        # Bivariate plots:
        profx = kws.pop("profx", False)
        profy = kws.pop("profy", False)

        if lower:
            for ax, axis in zip(self.offdiag_axs, self.offdiag_indices):
                if prof_edge_only:
                    if profx:
                        kws["profx"] = axis[1] == self.ndim - 1
                    if profy:
                        kws["profy"] = axis[0] == 0

                if kws["kind"] in ["hist", "contour", "contourf"]:
                    kws["bins"] = [
                        hists[axis[0]].edges,
                        hists[axis[1]].edges,
                    ]

                _plot_points(points[:, axis], ax=ax, **kws)

        if upper and not self.corner:
            for ax, axis in zip(self.offdiag_axs_u, self.offdiag_indices_u):
                if kws["kind"] in ["hist", "contour", "contourf"]:
                    kws["bins"] = [
                        hists[axis[0]].edges,
                        hists[axis[1]].edges,
                    ]

                _plot_points(points[:, axis], ax=ax, **kws)

        self._post_plot()

    def _post_plot(self) -> None:
        self.new = False
        self._fake_diag_yticks()
        self._force_non_negative_diag_ymin()


class SliceGrid:
    """Grid for slice matrix plots (https://arxiv.org/abs/2301.04178).

    This plot is used to visualize four dimensions of a distribution f(x1, x2, x3, x4).

    The main panel is an nrows x ncols grid that shows f(x1, x2 | x3, x4) -- the
    x1-x2 distribution for a planar slice in x3-x4. Each subplot corresponds to a
    different location in the x3-x4 plane.

    The following is only included if `marginals` is True:

        The bottom panel shows the marginal 3D distribution f(x1, x2 | x3).

        The right panel shows the marginal 3D distribution f(x1, x2 | x4).

        The bottom right subplot shows the full projection f(x1, x2).

        The lone subplot on the bottom right shows f(x1, x2)l, the full projection
        onto the x1-x2 plane.

    To do
    -----
    * Option to flow from top left to bottom right.

    Attributes
    ----------
    fig : proplot.figure.Figure
        The main figure.
    axs : proplot.gridspec.SubplotGrid
        The subplot axes.
    _axs : proplot.figure.Figure
        The subplot axes on the main panel.
    _axs_panel_x, _axs_panel_y, _axs_panel_xy : proplot.gridspec.SubplotGrid
        The subplot axes on the marginal panels.
    """

    def __init__(
        self,
        nrows: int = 9,
        ncols: int = 9,
        space: float = 0.0,
        gap: float = 2.0,
        marginals: bool = True,
        annotate: bool = True,
        annotate_kws_view: dict = None,
        annotate_kws_slice: dict = None,
        slice_label_height: float = 0.22,
        **fig_kws,
    ) -> None:
        """Constructor.

        nrows, ncols : int
            The number of rows/colums in the figure.
        space : float
            Spacing between subplots.
        gap : float
            Gap between main and marginal panels.
        marginals : bool
            Whether to include the marginal panels. If they are not included, we just
            have an nrows x ncols grid.
        annotate : bool
            Whether to add dimension labels/arrows to the figure.
        annotate_kws_view, annotate_kws_slice : dict
            Key word arguments for figure text. The 'view' key words are for the view
            dimension labels; they are printed on top of one of the subplots. The
            'slice' key words are for the slice dimension labels; they are printed
            on the sides of the figure, between the main and marginal panels.
        slice_label_height : float
            Tweaks the position of slice labels. Need a better way to handle this.
        **fig_kws
            Key word arguments for `uplt.subplots`.
        """
        self.nrows = nrows
        self.ncols = ncols
        self.space = space
        self.gap = gap
        self.marginals = marginals
        self.annotate = annotate
        self.slice_label_height = slice_label_height
        self.fig_kws = fig_kws
        self.axis_slice = None
        self.axis_view = None
        self.ind_slice = None

        self.annotate_kws_view = annotate_kws_view
        if self.annotate_kws_view is None:
            self.annotate_kws_view = {}
        self.annotate_kws_view.setdefault("color", "black")
        self.annotate_kws_view.setdefault("xycoords", "axes fraction")
        self.annotate_kws_view.setdefault("horizontalalignment", "center")
        self.annotate_kws_view.setdefault("verticalalignment", "center")

        self.annotate_kws_slice = annotate_kws_slice
        if self.annotate_kws_slice is None:
            self.annotate_kws_slice = {}
        self.annotate_kws_slice.setdefault("color", "black")
        self.annotate_kws_slice.setdefault("xycoords", "axes fraction")
        self.annotate_kws_slice.setdefault("horizontalalignment", "center")
        self.annotate_kws_slice.setdefault("verticalalignment", "center")
        self.annotate_kws_slice.setdefault(
            "arrowprops", dict(arrowstyle="->", color="black")
        )

        fig_kws.setdefault("figwidth", 8.5 * (ncols / 13.0))
        fig_kws.setdefault("share", True)
        fig_kws["ncols"] = ncols + 1 if marginals else ncols
        fig_kws["nrows"] = nrows + 1 if marginals else nrows
        hspace = nrows * [space]
        wspace = ncols * [space]
        if marginals:
            hspace[-1] = wspace[-1] = gap
        else:
            hspace = hspace[:-1]
            wspace = wspace[:-1]
        fig_kws["hspace"] = hspace
        fig_kws["wspace"] = wspace

        self.fig, self.axs = uplt.subplots(**fig_kws)

        self._axs = self.axs[:-1, :-1]
        self._axs_panel_x = []
        self._axs_panel_y = []
        if self.marginals:
            self._axs_panel_x = self.axs[-1, :]
            self._axs_panel_y = self.axs[:, -1]
        self._ax_panel_xy = self.axs[-1, -1]

    def _annotate(
        self,
        labels=None,
        slice_label_height=0.22,
        annotate_kws_view=None,
        annotate_kws_slice=None,
    ):
        """Add dimension labels and arrows."""
        # Label the view dimensions.
        for i, xy in enumerate([(0.5, 0.13), (0.12, 0.5)]):
            self.axs[0, 0].annotate(labels[i], xy=xy, **self.annotate_kws_view)

        # Label the slice dimensions. Print dimension labels with arrows like this:
        # "<----- x ----->" on the bottom and right side of the main panel.
        arrow_length = 2.5  # arrow length
        text_length = 0.15  # controls space between dimension label and start of arrow
        i = -1 - int(self.marginals)
        anchors = (
            self.axs[i, self.ncols // 2],
            self.axs[self.nrows // 2, i],
        )
        anchors[0].annotate(
            labels[2], xy=(0.5, -slice_label_height), **annotate_kws_slice
        )
        anchors[1].annotate(
            labels[3], xy=(1.0 + slice_label_height, 0.5), **annotate_kws_slice
        )
        for arrow_direction in (1.0, -1.0):
            anchors[0].annotate(
                "",
                xy=(0.5 + arrow_direction * arrow_length, -slice_label_height),
                xytext=(0.5 + arrow_direction * text_length, -slice_label_height),
                **annotate_kws_slice,
            )
            anchors[1].annotate(
                "",
                xy=(1.0 + slice_label_height, 0.5 + arrow_direction * arrow_length),
                xytext=(1.0 + slice_label_height, 0.5 + arrow_direction * text_length),
                **annotate_kws_slice,
            )

    def get_ind_slice(self) -> int:
        """Return slice indices from latest plot call."""
        return self.ind_slice

    def get_axis_slice(self) -> int:
        """Return slice axis from latest plot call."""
        return self.axis_slice

    def get_axis_view(self) -> tuple[int, ...]:
        """Return view axis from latest plot call."""
        return self.axis_view

    def set_limits(self, limits) -> None:
        """Set the plot limits."""
        for ax in self.axs:
            ax.format(xlim=limits[0], ylim=limits[1])

    def plot_hist(
        self,
        hist: Histogram,
        axis_view: tuple[int, int] = (0, 1),
        axis_slice: tuple[int, int] = (2, 3),
        labels: list[str] = None,
        pad: float = 0.0,
        debug: bool = False,
        **kws,
    ) -> None:
        """Plot a four-dimensional histogram.

        The first two dimensions are plotted as the last two are sliced.

        Parameters
        ----------
        hist: Histogram
            A four-dimensional histogram.
        axis_view, axis_slice : tuple[int, int]
        labels : list[str]
            Label for each dimension.
        pad : int, float, list
            This determines the start/stop indices along the sliced dimensions. If
            0, space the indices along axis `i` uniformly between 0 and `values.shape[i]`.
            Otherwise, add a padding equal to `int(pad[i] * values.shape[i])`. So, if
            the shape=10 and pad=0.1, we would start from 1 and end at 9.
        debug : bool
            Whether to print debugging messages.
        **kws
            Key word arguments pass to `plot.image.plot`
        """

        # Setup
        # -----------------------------------------------------------------------
        
        if hist.ndim < 4:
            raise ValueError(f"hist.ndim = {hist.ndim} < 4")

        self.axis_view = axis_view
        self.axis_slice = axis_slice

        # Compute 4D/3D/2D projections.
        _hist = hist.project(axis=(axis_view + axis_slice))
        _hist_x = hist.project(axis=(axis_view + axis_slice[:1]))
        _hist_y = hist.project(axis=(axis_view + axis_slice[1:]))
        _hist_xy = hist.project(axis=axis_view)

        # Get slice indices
        pad_factors = pad
        if type(pad) in [float, int]:
            pad_factors = len(axis_slice) * [pad_factors]

        ind_slice = []
        for axis, nsteps, pad_factor in zip(axis_slice, [self.nrows, self.ncols], pad_factors):
            lo = hist.shape[axis] * pad_factor
            lo = int(lo)
            hi = hist.shape[axis] - 1 - lo

            if (hi - lo) < nsteps:
                raise ValueError(f"values.shape[{i}] < number of slice indices requested.")
                
            if (hi - lo) == (nsteps - 1):
                ind_slice.append(
                    [int(i) for i in np.arange(nsteps)]
                )
            else:
                ind_slice.append(
                    [int(i) for i in np.linspace(lo, hi, nsteps)]
                )

        ind_slice = tuple(ind_slice)
        self.ind_slice = ind_slice

        if debug:
            print("debug slice indices:")
            for ind in ind_slice:
                print(ind)

        # Slice the 4D histogram.
        axis_view, axis_slice = (0, 1), (2, 3)
        for axis, ind in zip(axis_slice, ind_slice):
            _hist = _hist.slice(axis=axis, ind=ind)

        # Slice the 3D projections.
        _hist_x = _hist_x.slice(axis=2, ind=ind_slice[0])
        _hist_y = _hist_y.slice(axis=2, ind=ind_slice[1])

        # Scale all to unit maximum
        _hist.scale_max()
        _hist_x.scale_max()
        _hist_y.scale_max()
        _hist_xy.scale_max()

        if debug:
            print("debug _hist.shape =", _hist_x.shape)
            print("debug _hist_x.shape =", _hist_x.shape)
            print("debug _hist_y.shape =", _hist_y.shape)
            print("debug _hist_xy.shape =", _hist_xy.shape)

        # Get labels
        if labels is not None:
            labels = [labels[axis] for axis in axis_view + axis_slice]
            
        # Add dimension labels to the figure.
        if self.annotate and labels is not None:
            self._annotate(
                labels=labels,
                slice_label_height=self.slice_label_height,
                annotate_kws_view=self.annotate_kws_view,
                annotate_kws_slice=self.annotate_kws_slice,
            )


        # Plotting
        # -----------------------------------------------------------------------
        
        for i in range(self.nrows):
            for j in range(self.ncols):
                ax = self.axs[self.nrows - 1 - i, j]
                _hist_slice = _hist.slice(axis=axis_slice, ind=[j, i])
                _plot_hist(_hist_slice, ax=ax, **kws)

        if self.marginals:
            for i, ax in enumerate(reversed(self.axs[:-1, -1])):
                _hist_y_slice = _hist_y.slice(axis=2, ind=i)
                _plot_hist(_hist_y_slice, ax=ax, **kws)

            for i, ax in enumerate(self.axs[-1, :-1]):
                _hist_x_slice = _hist_x.slice(axis=2, ind=i)
                _plot_hist(_hist_x_slice, ax=ax, **kws)

            _plot_hist(_hist_xy, ax=self.axs[-1, -1], **kws)



import numpy as np
import scipy.optimize

from ..hist import Histogram
from ..hist import Histogram1D


def fit_linear(x: np.ndarray, y: np.ndarray) -> tuple[np.ndarray, float, float]:
    """Return (y_fit, slope, intercept) from linear fit."""

    def function(x: np.ndarray, slope: float, intercept: float) -> np.ndarray:
        return slope * x + intercept

    popt, pcov = scipy.optimize.curve_fit(function, x, y)
    slope, intercept = popt
    y_fit = function(x, slope, intercept)
    return (y_fit, slope, intercept)


def fit_normal(
    x: np.ndarray, y: np.ndarray
) -> tuple[np.ndarray, float, float, float, float]:
    """Return (yfit, sigma, mu, amplitude, offset) from Gaussian fit."""

    def function(x, sigma, mu, amplitude, offset):
        amplitude = amplitude / (sigma * np.sqrt(2.0 * np.pi))
        return offset + amplitude * np.exp(-0.5 * ((x - mu) / sigma) ** 2)

    popt, pcov = scipy.optimize.curve_fit(function, x, y)
    sigma, mu, amplitude, offset = popt
    y_fit = function(x, slope, intercept)
    return (y_fit, sigma, mu, amplitude, offset)



from . import cov
from . import density
from . import distributions
from . import hist
from . import core
from . import plot
from . import utils

from .cov import *
from .distributions import *
from .core import *
from .density import *
from .distance import *
from .hist import Grid
from .hist import Histogram
from .hist import Histogram1D
from .hist import SparseHistogram
from .hist import blur_hist
from .hist import clip_hist
from .hist import fill_hist
from .hist import thresh_hist
from .hist import sample_hist
from .hist import sample_sparse_hist
from .utils import *



"""Plotting routines for points."""
import numpy as np
import matplotlib.pyplot as plt
import ultraplot as uplt
from ipywidgets import interactive
from ipywidgets import widgets

from ..core import downsample as _downsample
from ..core import limits as get_limits
from ..core import combine_limits
from ..density import estimate_density as _estimate_density
from ..hist import Histogram
from ..hist import Histogram1D
from .. import core
from .core import plot_rms_ellipse_cov as _plot_rms_ellipse
from .hist import plot as _plot_hist
from .hist import plot_1d as _plot_hist_1d


def plot_hist_1d(
    points: np.ndarray,
    bins: int = 10,
    limits: np.ndarray = None,
    ax=None,
    **kws,
):
    """Plot one-dimensional histogram."""
    values, edges = np.histogram(points, bins=bins, range=limits)
    hist = Histogram1D(values, edges=edges)
    return _plot_hist_1d(hist, ax=ax, **kws)


def plot_density_estimate_1d(
    points: np.ndarray,
    bins: int = 10,
    limits: np.ndarray = None,
    method: str = "kde",
    method_kws: dict = None,
    ax=None,
    **kws,
):
    """Plot one-dimensional density estimate."""
    if method_kws is None:
        method_kws = {}

    if limits is None:
        limits = get_limits(points)

    edges = np.linspace(limits[0], limits[1], bins + 1)
    hist = Histogram1D(edges=edges)

    eval_points = hist.coords
    values = _estimate_density(points, eval_points, method, **method_kws)
    values = values.reshape(hist.shape)
    hist.values = values
    hist.normalize()
    return _plot_hist_1d(hist, ax=ax, **kws)


def plot_1d(points: np.ndarray, kde: bool = False, **kws):
    """Plot one-dimensional density."""
    plot_function = None
    if kde:
        plot_function = plot_density_estimate_1d
    else:
        plot_function = plot_hist_1d

    return plot_function(points, **kws)


def plot_rms_ellipse(
    points: np.ndarray,
    level: float | list[float] = 1.0,
    center_at_mean: bool = True,
    ax=None,
    **ellipse_kws,
):
    """Compute and plot RMS ellipse from bunch coordinates.

    Parameters
    ----------
    points : ndarray, shape (..., 2)
        Particle coordinates.
    level : number of list of numbers
        If a number, plot the rms ellipse inflated by the number. If a list of
        numbers, repeat for each number.
    center_at_mean : bool
        Whether to center the ellipse at the image centroid.
    """
    center = np.mean(points, axis=0)
    if not center_at_mean:
        center = (0.0, 0.0)
    return _plot_rms_ellipse(
        np.cov(points.T), center, level=level, ax=ax, **ellipse_kws
    )


def plot_scatter(points: np.ndarray, samples: int = None, ax=None, **kws):
    """Convenience function for 2D scatter plot.

    Parameters
    ----------
    points: np.ndarray, shape (..., n)
        Coordinate array for n points in d-dimensional space.
    samples : int
        Plot this many random samples.
    **kws
        Key word arguments passed to `ax.scatter`.
    """
    if "color" in kws:
        kws["c"] = kws.pop("color")

    for kw in ["size", "ms"]:
        if kw in kws:
            kws["s"] = kws.pop(kw)

    kws.setdefault("c", "black")
    kws.setdefault("ec", "None")
    kws.setdefault("s", 2.0)

    if samples is not None:
        points = _downsample(points, samples)
    return ax.scatter(points[:, 0], points[:, 1], **kws)


def plot_hist(
    points: np.ndarray,
    bins: int = 50,
    limits: np.ndarray = None,
    ax=None,
    **kws,
):
    """Plot two-dimensional histogram.

    Parameters
    ----------
    points: np.ndarray, shape (..., n)
        Particle coordinates.
    limits, bins :
        See `psdist.bunch.histogram`.
    **kws
        Key word arguments passed to `plotting.image`.
    """
    if bins is None:
        bins = 50

    values, edges = np.histogramdd(points, bins=bins, range=limits)
    hist = Histogram(values, edges=edges)
    return _plot_hist(hist, ax=ax, **kws)


def plot_density_estimate(
    points: np.ndarray,
    bins: int = 10,
    limits: np.ndarray = None,
    method: str = "kde",
    method_kws: dict = None,
    ax=None,
    **kws,
):
    """Plot two-dimensional density estimate.

    Parameters
    ----------
    points: np.ndarray, shape (..., N)
        Particle coordinates.
    bins : int
       Number of bins defining the evaluation grid along each dimension.
    limits : np.ndarray
        Evaluation grid limits [(xmin, xmax), (ymin, ymax)].
    method: str
        Name of density estimator.
    method_kws: dict
        Key word arguments passed to density estimator.
    **kws
        Key word arguments passed to `plot_hist`.
    """
    if method_kws is None:
        method_kws = {}

    if limits is None:
        limits = get_limits(points)

    ndim = points.shape[1]

    edges = [np.linspace(limits[i][0], limits[i][1], bins + 1) for i in range(ndim)]
    hist = Histogram(edges=edges)
    eval_points = hist.points()

    values = _estimate_density(points, eval_points, method, **method_kws)
    values = values.reshape(hist.shape)
    hist.values = values
    hist.normalize()
    return _plot_hist(hist, ax=ax, **kws)


def plot(
    points: np.ndarray,
    kind: str = "hist",
    rms_ellipse: bool = False,
    rms_ellipse_kws: dict = None,
    ax=None,
    **kws,
):
    """Plot two-dimensional density.

    Parameters
    ----------
    points: np.ndarray, shape (..., N)
        Coordinates of points in N-dimensional space.
    kind : {'hist', 'contour', 'contourf', 'scatter', 'kde'}
        The kind of plot. These key words map to the following functions:
        "hist": psdist.plot.points.plot_hist(kind="hist")
        "contour": psdist.plot.points.plot_hist(kind="contour")
        "contourf": psdist.plot.points.plot_hist(kind="contourf")
        "scatter": psdist.plot.points.plot_scatter
        "density": psdist.plot.points.plot_density_estimate
    ax : Axes
        The axis on which to plot.
    **kws
        Key word arguments passed to plotting function.
    """
    if kind == "hist":
        kws.setdefault("mask", True)

    if rms_ellipse_kws is None:
        rms_ellipse_kws = {}

    plot_function = None
    if kind in ["hist", "contour", "contourf"]:
        plot_function = plot_hist
        if kind == "hist":
            kws["kind"] = "pcolor"
        if kind == "contour":
            kws["kind"] = "contour"
        if kind == "contourf":
            kws["kind"] = "contourf"
    elif kind == "scatter":
        plot_function = plot_scatter
    elif kind == "density":
        plot_function = plot_density_estimate
    else:
        raise ValueError(f"Invalid plot kind '{kind}'.")

    output = plot_function(points, ax=ax, **kws)
    if rms_ellipse:
        plot_rms_ellipse(points, ax=ax, **rms_ellipse_kws)
    return output


def plot_joint(points: np.ndarray, grid_kws: dict = None, **kws):
    from psdist.plot.grid import JointGrid

    if grid_kws is None:
        grid_kws = {}

    grid = JointGrid(**grid_kws)
    grid.plot_points(points, **kws)
    return grid


def plot_corner(points: np.ndarray, grid_kws: dict = None, **kws):
    from psdist.plot.grid import CornerGrid

    if grid_kws is None:
        grid_kws = {}

    ndim = points.shape[1]

    grid = CornerGrid(ndim=ndim, **grid_kws)
    grid.plot_points(points, **kws)
    return grid


def plot_interactive_slice_2d(
    data: np.ndarray | list[np.ndarray] | list[list[np.ndarray]],
    limits: list[tuple[float, float]] = None,
    share_limits: int = 1,
    default_axis: tuple[int] = (0, 1),
    slice_type: str = "int",
    plot_res: int = 64,
    slice_res: int = 16,
    dims: list[str] = None,
    units: list[str] = None,
    options: dict = None,
    autolim_kws: dict = None,
    fig_kws: dict = None,
    **plot_kws,
):
    """Plot two-dimensional density with interactive slicing.

    Parameters
    ----------
    data : ndarray, shape (..., N) or list[ndarray] or list[list[ndarray]]
        ndarray:
            Corresponds to a single collection (bunch) of particles.
        list[ndarray]:
            Corresponds to a series of bunches. A widget is added to select the
            bunch to plot.
        list[list[ndarray]]:
            If there are K different lists provided, a K-column figure is produced.
            For example, if we had two different simulations, each generating a
            series of bunches at the same time steps, the function would create
            two subplots and a widget to select the frame.
    limits : list[tuple[float, float]]
        Limits (xmin, xmax) along each axis.
    share_limits : 0, 1, 2
        Whether to share axis limits across frames. If 0, don't share. If 1, share
        between each subplot for each figure (frame). If 2, share for all
        figures/subplots.
    default_axis : (int, int)
        Default view axis.
    slice_type : {"int", "range"}
        Whether to slice one index along the axis or a range of indices.
    plot_res, slice_res : int
        Default grid resolution for plotting/slicing. These can be updated using
        the interactive widgets.
    dims, units : list[str], shape (n,)
        Dimension names and units.
    options : dict
        Determines the widgets to be displayed. Options are:
        - "auto_plot_res": Option to set bins="auto" in histogram.
        - "discrete": Option for discrete colormap norm. (Default: False).
        - "ellipse": Option to plot rms ellipse. (Default: False)
        - "log": Option for logarithmic colormap scaling. (Default: True)
        - "mask": Option to include a small offset so that there are no zero bins. (Default: False)
        - "normalize": Option to normalize x-px, y-py, z-pz to unit covariance matrix. (Default: False)
        - "profiles": Option to plot profiles (line-outs) on bottom and left spines.
    autolim_kws : dict
        Key word arguments passed to `auto_limits`.
    fig_kws : dict
        Key word arguments passed to `proplot.subplots`.
    **plot_kws
        Key word arguments passed to `plot`.
    """
    if type(data) is not list:
        data = [data]
        
    if type(data[0]) is not list:
        data = [data]

    nrows = len(data)
    ncols = len(data[0])
    ndims = data[0][0].shape[1]

    for i in range(nrows):
        if len(data[i]) != ncols:
            raise ValueError("lists must have same length")

    for i in range(nrows):
        for j in range(ncols):
            if data[i][j].shape[1] != ndims:
                raise ValueError("data must have the same number of dimensions.")

    if fig_kws is None:
        fig_kws = {}

    plot_kws.setdefault("kind", "hist")
    plot_kws.setdefault(
        "rms_ellipse_kws",
        {
            "level": [1.0, 2.0, 3.0, 4.0, 5.0],
            "color": "white",
            "alpha": 0.2,
            "lw": 0.4,
        },
    )

    # Compute limits [(xmin, xmax), ...] for each bunch (data[i][j]).
    if autolim_kws is None:
        autolim_kws = {}
        
    if limits is None:
        limits_list = np.zeros((nrows, ncols, ndims, 2))
        for i in range(nrows):
            for j in range(ncols):
                limits_list[i, j, :, :] = get_limits(data[i][j], **autolim_kws)
        if share_limits == 1:
            for j in range(ncols):
                limits = combine_limits(limits_list[:, j, :, :])
                for i in range(nrows):
                    limits_list[i, j] = limits
        elif share_limits == 2:
            limits = combine_limits(
                limits_list.reshape((nrows * ncols, ndims, 2))
            )
            for i in range(nrows):
                for j in range(ncols):
                    limits_list[i, j] = limits
    else:
        limits_list = [[limits for _ in range(ncols)] for _ in range(nrows)]

    # Set axis labels.
    if dims is None:
        dims = [f"x{i + 1}" for i in range(ndims)]
    if units is None:
        units = ndims * [""]
    dims_units = []
    for dim, unit in zip(dims, units):
        dims_units.append(f"{dim}" + f" [{unit}]" if unit != "" else dim)

    # Widgets
    _widgets = {}
    _widgets["dim1"] = widgets.Dropdown(
        options=dims,
        index=default_axis[0],
        description="dim 1"
    )
    _widgets["dim2"] = widgets.Dropdown(
        options=dims,
        index=default_axis[1],
        description="dim 2"
    )
    _widgets["frame"] = widgets.BoundedIntText(
        min=0,
        max=(ncols - 1),
        description="frame"
    )
    _widgets["slice_res"] = widgets.BoundedIntText(
        value=slice_res,
        min=2,
        max=200,
        step=1,
        description="slice_res",
    )
    _widgets["plot_res"] = widgets.BoundedIntText(
        value=plot_res,
        min=2,
        max=350,
        step=1,
        description="plot_res",
    )
    _widgets["auto_plot_res"] = widgets.Checkbox(
        description="auto_plot_res", value=False
    )
    _widgets["discrete"] = widgets.Checkbox(description="discrete", value=False)
    _widgets["ellipse"] = widgets.Checkbox(description="ellipse", value=False)
    _widgets["log"] = widgets.Checkbox(description="log", value=False)
    _widgets["mask"] = widgets.Checkbox(description="mask", value=False)
    _widgets["normalize"] = widgets.Checkbox(description="normalize", value=False)
    _widgets["profiles"] = widgets.Checkbox(description="profiles", value=False)

    # Sliders and checkboxes for slicing:
    _widgets["sliders"] = []
    _widgets["checks"] = []
    for k in range(ndims):
        if slice_type == "int":
            slider = widgets.IntSlider(
                min=0,
                max=(_widgets["slice_res"].value - 1),
                value=int(_widgets["slice_res"].value / 2),
                description=dims[k],
                continuous_update=True,
            )
        elif slice_type == "range":
            slider = widgets.IntRangeSlider(
                min=0,
                max=(_widgets["slice_res"].value - 1),
                value=(0, _widgets["slice_res"].value - 1),
                description=dims[k],
                continuous_update=True,
            )
        else:
            raise ValueError("Invalid `slice_type`.")
        slider.layout.display = "none"
        _widgets["sliders"].append(slider)
        _widgets["checks"].append(widgets.Checkbox(description=f"slice {dims[k]}"))

    def hide(button):
        """Hide inactive sliders."""
        for k in range(ndims):
            # Hide elements for dimensions being plotted.
            valid = dims[k] not in [_widgets["dim1"].value, _widgets["dim2"].value]
            disp = None if valid else "none"
            for element in [_widgets["sliders"][k], _widgets["checks"][k]]:
                element.layout.display = disp

            # Uncheck boxes for dimensions being plotted.
            if not valid and _widgets["checks"][k].value:
                _widgets["checks"][k].value = False

            # Make sliders respond to check boxes.
            if not _widgets["checks"][k].value:
                _widgets["sliders"][k].layout.display = "none"
            _widgets["plot_res"].layout.display = (
                "none" if _widgets["auto_plot_res"].value else None
            )

    # Make slider visibility depend on checkmarks.
    for element in (
        _widgets["dim1"],
        _widgets["dim2"],
        *_widgets["checks"],
        _widgets["auto_plot_res"],
    ):
        element.observe(hide, names="value")

    # Initial hide
    for k in range(ndims):
        if k in default_axis:
            _widgets["checks"][k].layout.display = "none"
            _widgets["sliders"][k].layout.display = "none"
    if ncols == 1:
        _widgets["frame"].layout.display = "none"

    # Set default options.
    if options is None:
        options = {}
    options.setdefault("auto_plot_res", False)
    options.setdefault("discrete", False)
    options.setdefault("ellipse", False)
    options.setdefault("log", True)
    options.setdefault("mask", False)
    options.setdefault("normalize", False)
    options.setdefault("profiles", False)

    # Show/hide widgets based on `options`.
    for name, setting in options.items():
        _widgets[name].layout.display = None if setting else "none"

    plot_kws.setdefault("offset", 0.0)
    default_offset = plot_kws["offset"]


    def update(**kws):
        # Collect key word arguments.
        frame = kws["frame"]
        dim1 = kws["dim1"]
        dim2 = kws["dim2"]
        slice_res = kws["slice_res"]
        plot_res = kws["plot_res"]
        auto_plot_res = kws["auto_plot_res"]

        # Update the slider ranges/values based on slice_res.
        for slider in _widgets["sliders"]:
            slider.max = slice_res - 1

        # Collect slice indices.
        ind, checks = [], []
        for i in range(1, ndims + 1):
            if f"check{i}" in kws:
                checks.append(kws[f"check{i}"])
            if f"slider{i}" in kws:
                _ind = kws[f"slider{i}"]
                if type(_ind) is int:
                    _ind = (_ind, _ind + 1)
                ind.append(_ind)

        # Exit if input is invalid.
        for dim, check in zip(dims, checks):
            if check and dim in (dim1, dim2):
                return
        if dim1 == dim2:
            return

        # Collect data
        _data = [data[index][frame] for index in range(nrows)]
        _limits_list = limits_list[:, frame, :, :]

        # Normalize coordinates
        if kws["normalize"]:
            for i, _points in enumerate(_data):
                if _points.shape[1] % 2 == 0:
                    _data[i] = core.normalize_2d_projections(_points, scale=True)

            _limits_list = [core.limits(_points, **autolim_kws) for _points in _data]
            if share_limits > 0:
                _limits_list = [
                    combine_limits(_limits_list) for _ in range(len(_data))
                ]

        # Slice
        axis_view = [dims.index(dim) for dim in (dim1, dim2)]
        axis_slice = [dims.index(dim) for dim, check in zip(dims, checks) if check]
        if axis_slice:
            for index in range(nrows):
                slice_limits = []
                for k in axis_slice:
                    (imin, imax) = ind[k]
                    (xmin, xmax) = _limits_list[index][k]
                    edges = np.linspace(xmin, xmax, slice_res + 1)

                    if imax > len(edges) - 1:
                        print(f"{dims[k]} out of range.")
                        return

                    slice_limits.append((edges[imin], edges[imax]))

                _data[index] = core.slice_(_data[index], axis=axis_slice, limits=slice_limits)

        # Handle empty slice (do nothing).
        for _points in _data:
            if _points.shape[0] == 0:
                return

        # Update plotting key word arguments.
        if plot_kws["kind"] != "scatter":
            plot_kws["bins"] = "auto" if auto_plot_res else plot_res
            plot_kws["discrete"] = kws["discrete"]
            plot_kws["mask"] = kws["mask"]
            plot_kws["norm"] = "log" if kws["log"] else None
            plot_kws["profx"] = kws["profiles"]
            plot_kws["profy"] = kws["profiles"]
            plot_kws["rms_ellipse"] = kws["ellipse"]

            # Add a small offset to the image if we are not masking and are using a logarithmic colormap.
            if plot_kws["norm"] == "log" and not plot_kws["mask"]:
                if default_offset > 0.0:
                    plot_kws["offset"] = default_offset
                else:
                    plot_kws["offset"] = 1.0
            else:
                plot_kws["offset"] = default_offset

            # Temporary bug fix: If we check and then uncheck "log", and
            # the colorbar has minor ticks, the tick label formatter will
            # remain in "log" mode forever after.
            if "colorbar_kw" in plot_kws:
                if "tickminor" in plot_kws["colorbar_kw"] and not kws["log"]:
                    plot_kws["colorbar_kw"]["formatter"] = None

        ## Create figure.
        fig, axs = uplt.subplots(
            ncols=nrows,
            sharex=(share_limits and nrows > 1),
            sharey=(share_limits and nrows > 1),
            **fig_kws,
        )
        for index, ax in enumerate(axs):
            limits = [_limits_list[index][k] for k in axis_view]
            if plot_kws["kind"] != "scatter":
                plot_kws["limits"] = limits

            plot(_data[index][:, axis_view], ax=ax, **plot_kws)

            ax.format(xlim=limits[0], ylim=limits[1])

        labels = dims if _widgets["normalize"].value else dims_units
        axs.format(
            xlabel=labels[axis_view[0]],
            ylabel=labels[axis_view[1]],
        )
        plt.show()

    # Pass key word arguments to `ipywidgets.interactive`.
    kws = {}
    for key in [
        "frame",
        "dim1",
        "dim2",
        "slice_res",
        "plot_res",
        "auto_plot_res",
        "discrete",
        "ellipse",
        "log",
        "mask",
        "normalize",
        "profiles",
    ]:
        kws[key] = _widgets[key]
    for i, check in enumerate(_widgets["checks"], start=1):
        kws[f"check{i}"] = check
    for i, slider in enumerate(_widgets["sliders"], start=1):
        kws[f"slider{i}"] = slider
    return interactive(update, **kws)


def plot_interactive_slice_1d(
    data: np.ndarray | list[np.ndarray] | list[list[np.ndarray]],
    limits: list[tuple[float]] = None,
    share_limits: int = 1,
    default_axis: int = 0,
    slice_type: str = "int",
    plot_res: int = 64,
    slice_res: int = 16,
    dims: list[str] = None,
    units: list[str] = None,
    options: dict = None,
    colors: list[str] = None,
    cycle: str = None,
    labels: list[str] = None,
    legend: bool = False,
    update_limits_on_slice: bool = False,
    legend_kws: dict = None,
    autolim_kws: dict = None,
    fig_kws: dict = None,
    **plot_kws,
):
    """One-dimensional partial projection of one or more bunches (or series of bunches)
    with interactive slicing.

    Profiles are scaled to unit area by default.

    Parameters
    ----------
    data : ndarray, shape (n, d) or list[ndarray] or list[list[ndarray]]
        - Particle coordinates.
        - List of L bunches: generates widget to select the frame to plot.
        - K lists of L bunches: generates K-column figure with widget to select one
          of the L frames.
          Example: Compare the evolution of K=3 bunches at L=6 frames.
    limits : list[(min, max)]
        Limits along each axis.
    share_limits : bool
        Whether to share axis limits across frames.
    default_axis : int
        Default view axis.
    slice_type : {"int", "range"}
        Whether to slice one index along the axis or a range of indices.
    plot_res, slice_res : int
        Default grid resolution for plotting/slicing. These can be updated using
        the interactive widgets.
    dims, units : list[str], shape (n,)
        Dimension names and units.
    colors : list
        List of colors, otherwise use default color cycle.
    cycle : str
        If the name of a cycle is passed, override `colors`.
    options : dict
        Determines the widgets to be displayed. Options are:
        - "alpha": Option to scale plot opacity (alpha). (Default: False)
        - "auto_plot_res": Option to set bins="auto" in histogram. (Default: False)
        - "log": Option for logarithmic y axis scale. (Default: True)
        - "normalize": Option to normalize x-px, y-py, z-pz to unit covariance matrix. (Default: False)
        - "scale": Option to scale to unit max or unit area. (Default: False)
    labels : list[str]
        Labels for legend.
    legend : bool
        Whether to include legend.
    update_limits_on_slice : bool
        Recompute limits when the distribution is sliced. Otherwise, keep the limits
        from the full distribution.
    legend_kws : bool
        Key word arguments for legend.
    autolim_kws : dict
        Key word arguments passed to `auto_limits`.
    fig_kws : dict
        Key word arguments passed to `proplot.subplots`.
    **plot_kws
        Key word arguments passed to `plot_profile`.
    """
    if type(data) is not list:
        data = [data]

    if type(data[0]) is not list:
        data = [data]

    nrows = len(data)
    ncols = len(data[0])
    ndims = data[0][0].shape[1]

    for i in range(nrows):
        if len(data[i]) != ncols:
            raise ValueError("lists must have same length")

    for i in range(nrows):
        for j in range(ncols):
            if data[i][j].shape[1] != ndims:
                raise ValueError("data must have the same number of dimensions.")

    if fig_kws is None:
        fig_kws = {}

    # Plot color cycle.
    if colors is None:
        if cycle is None:
            colors = uplt.Cycle(uplt.rc["cycle"]).by_key()["color"]
        else:
            colors = uplt.Cycle(cycle).by_key()["color"]
    if nrows == 1:
        for key in ["color", "c"]:
            if key in plot_kws:
                colors = ncols * [plot_kws.pop(key)]

    # Legend
    if legend_kws is None:
        legend_kws = {}
    legend_kws.setdefault("loc", "right")
    legend_kws.setdefault("framealpha", 0.0)
    legend_kws.setdefault("ncols", 1)
    
    if labels is None:
        labels = [f"prof_{i}" for i in range(nrows)]

    # Compute limits [(xmin, xmax), ...] for each frame.
    if autolim_kws is None:
        autolim_kws = {}
    if limits is None:
        limits_list = np.zeros((ncols, ndims, 2))
        for j in range(ncols):
            limits_list[j] = combine_limits(
                [core.limits(data[i][j], **autolim_kws) for i in range(nrows)]
            )
        if share_limits:
            limits = combine_limits(limits_list)
            for j in range(ncols):
                limits_list[j] = limits
    else:
        limits_list = [[limits for _ in range(ncols)] for _ in range(nrows)]

    # Set axis labels.
    if dims is None:
        dims = [f"x{i + 1}" for i in range(ndims)]
    if units is None:
        units = ndims * [""]
    dims_units = []
    for dim, unit in zip(dims, units):
        dims_units.append(f"{dim}" + f" [{unit}]" if unit != "" else dim)

    # Widgets
    _widgets = {}
    _widgets["dim"] = widgets.Dropdown(
        options=dims, index=default_axis, description="dim"
    )
    _widgets["frame"] = widgets.BoundedIntText(
        min=0, max=(ncols - 1), description="frame"
    )
    _widgets["slice_res"] = widgets.BoundedIntText(
        value=slice_res,
        min=2,
        max=200,
        step=1,
        description="slice_res",
    )
    _widgets["plot_res"] = widgets.BoundedIntText(
        value=plot_res,
        min=2,
        max=350,
        step=1,
        description="plot_res",
    )
    _widgets["alpha"] = widgets.FloatSlider(
        min=0.0, max=1.0, value=1.0, description="alpha"
    )
    _widgets["auto_plot_res"] = widgets.Checkbox(
        description="auto_plot_res", value=False
    )
    _widgets["kind"] = widgets.Dropdown(
        description="kind",
        options=["step", "stepfilled", "line", "linefilled"],
        value="step",
    )
    _widgets["kde"] = widgets.Checkbox(desciption="kde", value=False)
    _widgets["log"] = widgets.Checkbox(description="log", value=False)
    _widgets["normalize"] = widgets.Checkbox(description="normalize", value=False)
    _widgets["scale"] = widgets.Dropdown(
        options=["none", "density", "max"], description="scale", value="density"
    )

    # Sliders and checkboxes for slicing:
    _widgets["sliders"] = []
    _widgets["checks"] = []
    for k in range(ndims):
        if slice_type == "int":
            slider = widgets.IntSlider(
                min=0,
                max=(_widgets["slice_res"].value - 1),
                value=int(_widgets["slice_res"].value / 2),
                description=dims[k],
                continuous_update=True,
            )
        elif slice_type == "range":
            slider = widgets.IntRangeSlider(
                min=0,
                max=(_widgets["slice_res"].value - 1),
                value=(0, _widgets["slice_res"].value - 1),
                description=dims[k],
                continuous_update=True,
            )
        else:
            raise ValueError("Invalid `slice_type`.")
        slider.layout.display = "none"
        _widgets["sliders"].append(slider)
        _widgets["checks"].append(widgets.Checkbox(description=f"slice {dims[k]}"))

    def hide(button):
        """Hide inactive sliders."""
        for k in range(ndims):
            # Hide elements for dimensions being plotted.
            valid = dims[k] != _widgets["dim"].value
            disp = None if valid else "none"
            for element in [_widgets["sliders"][k], _widgets["checks"][k]]:
                element.layout.display = disp
                
            # Uncheck boxes for dimensions being plotted.
            if not valid and _widgets["checks"][k].value:
                _widgets["checks"][k].value = False
                
            # Make sliders respond to check boxes.
            if not _widgets["checks"][k].value:
                _widgets["sliders"][k].layout.display = "none"
                
            _widgets["plot_res"].layout.display = (
                "none" if _widgets["auto_plot_res"].value else None
            )

    # Make slider visibility depend on checkmarks.
    for element in [_widgets["dim"], *_widgets["checks"], _widgets["auto_plot_res"]]:
        element.observe(hide, names="value")

    # Initial hide
    for k in range(ndims):
        if k == default_axis:
            _widgets["checks"][k].layout.display = "none"
            _widgets["sliders"][k].layout.display = "none"
    if ncols == 1:
        _widgets["frame"].layout.display = "none"

    # Set default options.
    if options is None:
        options = {}
    options.setdefault("alpha", False)
    options.setdefault("auto_plot_res", False)
    options.setdefault("kde", False)
    options.setdefault("kind", False)
    options.setdefault("scale", False)
    options.setdefault("log", True)
    options.setdefault("normalize", False)

    # Show/hide widgets based on `options`.
    for name, setting in options.items():
        _widgets[name].layout.display = None if setting else "none"

    def update(**kws):
        # Collect key word arguments.
        frame = kws["frame"]
        dim_view = kws["dim"]

        # Update the slider ranges/values based on slice_res.
        for slider in _widgets["sliders"]:
            slider.max = kws["slice_res"] - 1

        # Collect slice indices.
        ind, checks = [], []
        for i in range(1, ndims + 1):
            if f"check{i}" in kws:
                checks.append(kws[f"check{i}"])
            if f"slider{i}" in kws:
                _ind = kws[f"slider{i}"]
                if type(_ind) is int:
                    _ind = (_ind, _ind + 1)
                ind.append(_ind)

        # Exit if input is invalid.
        for dim, check in zip(dims, checks):
            if check and dim == dim_view:
                return

        # Collect data.
        _data = [data[index][frame] for index in range(nrows)]
        limits = limits_list[frame, :, :]

        # Normalize coordinates.
        if kws["normalize"]:
            for i, _points in enumerate(_data):
                if _points.shape[1] % 2 == 0:
                    _data[i] = core.normalize_2d_projections(_points, scale=True)

            limits = [core.limits(_points, **autolim_kws) for _points in _data]
            limits = combine_limits(limits)

        # Slice
        axis_view = dims.index(dim_view)
        axis_slice = [dims.index(dim) for dim, check in zip(dims, checks) if check]
        if axis_slice:
            for index in range(nrows):
                slice_limits = []
                for k in axis_slice:
                    (imin, imax) = ind[k]
                    (xmin, xmax) = limits[k]
                    edges = np.linspace(xmin, xmax, kws["slice_res"] + 1)
                    if imax > len(edges) - 1:
                        print(f"{dims[k]} out of range.")
                        return
                    slice_limits.append((edges[imin], edges[imax]))

                _data[index] = core.slice_(
                    _data[index], axis=axis_slice, limits=slice_limits
                )

        # Handle empty slice (do nothing).
        for _points in _data:
            if _points.shape[0] == 0:
                return

        # Recompute limits from sliced data.
        if update_limits_on_slice:
            limits = combine_limits(
                [core.limits(_points, **autolim_kws) for _points in _data]
            )

        # Plot
        fig, ax = uplt.subplots(**fig_kws)
        for index in range(len(_data)):
            plot_kws["color"] = colors[index]
            plot_kws["label"] = labels[index]
            plot_kws["alpha"] = kws["alpha"]
            plot_kws["kind"] = kws["kind"]
            plot_kws["scale"] = kws["scale"]
            plot_kws["kde"] = kws["kde"]

            bins = "auto" if kws["auto_plot_res"] else kws["plot_res"]

            x = _data[index][:, axis_view]
            plot_1d(x, bins=bins, limits=limits[axis_view], ax=ax, **plot_kws)

        if kws["log"]:
            ax.format(yscale="log", yformatter="log")

        ax.format(
            xlim=limits[axis_view],
            xlabel=(
                dims[axis_view]
                if _widgets["normalize"].value
                else dims_units[axis_view]
            ),
        )
        if legend and (labels is not None):
            ax.legend(**legend_kws)
        plt.show()

    # Pass key word arguments to `ipywidgets.interactive`.
    kws = {}
    for key in [
        "frame",
        "dim",
        "slice_res",
        "plot_res",
        "alpha",
        "auto_plot_res",
        "kind",
        "kde",
        "log",
        "normalize",
        "scale",
    ]:
        kws[key] = _widgets[key]
    for i, check in enumerate(_widgets["checks"], start=1):
        kws[f"check{i}"] = check
    for i, slider in enumerate(_widgets["sliders"], start=1):
        kws[f"slider{i}"] = slider
    return interactive(update, **kws)



import numpy as np
import psdist as ps


def test_project():
    shape = (6, 4, 12, 8, 2, 9)
    values = np.abs(np.random.normal(size=shape))
    coords = [np.arange(s) for s in shape]

    hist = ps.hist.Histogram(values, coords)

    for axis in np.ndindex(*(values.ndim * [values.ndim])):
        if len(np.unique(axis)) != values.ndim:
            continue
        hist_proj = ps.hist.project(hist, axis)
        shape = hist_proj.shape
        correct_shape = tuple([values.shape[i] for i in axis])
        assert shape == correct_shape


def test_slice_idx():
    shape = (6, 4, 12, 8, 2, 9)
    values = np.abs(np.random.normal(size=shape))
    axis = (2, 0, 3, 5)
    ind = (3, (3, 9), [4, 5, 6], 1)
    idx = ps.hist.slice_idx(values.ndim, axis=axis, ind=ind)
    values[idx]



import os
import pathlib
import numpy as np
import matplotlib.pyplot as plt
import ultraplot as uplt

import psdist as ps
import psdist.plot as psv


uplt.rc["cmap.discrete"] = False
uplt.rc["cmap.sequential"] = "viridis"
uplt.rc["grid"] = False
uplt.rc["savefig.dpi"] = 200


path = pathlib.Path(__file__)
output_dir = os.path.join("outputs", path.stem)
os.makedirs(output_dir, exist_ok=True)


def test_plot_ellipse():
    fig, ax = uplt.subplots()
    psv.plot_ellipse(r1=1.5, r2=0.5, ax=ax)
    ax.format(xlim=(-2.0, 2.0), ylim=(-2.0, 2.0))
    plt.savefig(os.path.join(output_dir, "fig_plot_ellipse.png"))
    plt.close()


def test_plot_circle():
    fig, ax = uplt.subplots()
    psv.plot_circle(ax=ax)
    ax.format(xlim=(-2.0, 2.0), ylim=(-2.0, 2.0))
    plt.savefig(os.path.join(output_dir, "fig_plot_circle.png"))


def test_plot_rms_ellipse_cov():
    fig, ax = uplt.subplots()
    psv.plot_rms_ellipse_cov(cov_matrix=np.eye(2), level=[0.5, 1.0], ax=ax)
    ax.format(xlim=(-2.0, 2.0), ylim=(-2.0, 2.0))
    plt.savefig(os.path.join(output_dir, "fig_plot_rms_ellipse.png"))
    plt.close()


def test_plot_points_rms_ellipse():
    x = np.random.normal(size=(1000, 2))
    
    fig, ax = uplt.subplots()
    psv.plot(x, kind="scatter", color="black", ax=ax)
    psv.plot_rms_ellipse(x, level=[0.5, 1.0], color="red", ax=ax)
    ax.format(xlim=(-2.0, 2.0), ylim=(-2.0, 2.0))
    plt.savefig(os.path.join(output_dir, "fig_plot_rms_ellipse.png"))
    plt.close()


def test_plot_hist_1d():
    x = np.random.normal(size=10_000)
    bin_edges = np.linspace(-4.0, 4.0, 51)
    hist = ps.Histogram1D(edges=bin_edges)
    hist.bin(x)

    fig, ax = uplt.subplots(figsize=(3.0, 1.5))
    psv.plot_hist_1d(hist, kind="step", ax=ax, color="black")

    plt.savefig(os.path.join(output_dir, "fig_plot_profile.png"))
    plt.close()


def test_plot_hist():
    x = np.random.normal(size=(10_000, 2))
    bin_edges = [np.linspace(-4.0, 4.0, 51), np.linspace(-4.0, 4.0, 51)]
    hist = ps.Histogram(edges=bin_edges)
    hist.bin(x)

    fig, ax = uplt.subplots(figwidth=3.0)
    psv.plot_hist(hist, ax=ax)

    plt.savefig(os.path.join(output_dir, "fig_plot_hist.png"))
    plt.close()


def test_plot_hist_prof():
    x = np.random.normal(size=(10_000, 2))
    bin_edges = [np.linspace(-4.0, 4.0, 51), np.linspace(-4.0, 4.0, 51)]
    hist = ps.Histogram(edges=bin_edges)
    hist.bin(x)

    fig, ax = uplt.subplots(figwidth=3.0)
    psv.plot_hist(hist, ax=ax, profx=True, profy=True)
    plt.savefig(os.path.join(output_dir, "fit_plot_hist_prof.png"))
    plt.close()


def test_plot_hist_rms_ellipse():
    x = np.random.normal(size=(10_000, 2))
    bin_edges = [np.linspace(-4.0, 4.0, 51), np.linspace(-4.0, 4.0, 51)]
    hist = ps.Histogram(edges=bin_edges)
    hist.bin(x)

    fig, ax = uplt.subplots(figwidth=3.0)
    psv.plot_hist(hist, ax=ax, rms_ellipse=True, rms_ellipse_kws=dict(level=[1.0, 2.0]))
    plt.savefig(os.path.join(output_dir, "fig_plot_hist_rms_ellipse.png"))
    plt.close()


def test_plot_hist():
    x = np.random.normal(size=(10_000, 2))
    bin_edges = [np.linspace(-4.0, 4.0, 51), np.linspace(-4.0, 4.0, 51)]

    fig, ax = uplt.subplots(figwidth=3.0)
    psv.plot(x, bins=bin_edges, kind="hist", ax=ax)
    plt.savefig(os.path.join(output_dir, "fig_plot_hist.png"))
    plt.close()


def test_plot_scatter():
    x = np.random.normal(size=(10_000, 2))

    fig, ax = uplt.subplots(figwidth=3.0)
    psv.plot(x, kind="scatter", ax=ax)
    plt.savefig(os.path.join(output_dir, "fig_plot_scatter.png"))
    plt.close()


def test_grid():
    pass



import numpy as np
import psdist as ps
from psdist import GaussianKDE


def gen_dist(ndim: int) -> np.ndarray:
    return np.random.rand(100, ndim)


def test_sparse_histogram():
    n_points = int(1.00e04)
    n_dims = 4
    n_bins = 10
    rng = np.random.default_rng(1234)

    x = rng.normal(size=(n_points, n_dims))

    edges = ps.histogram_bin_edges(x, bins=n_bins)
    sparse_hist = ps.sparse_histogram(x, bins=edges)

    hist = ps.histogram(x, bins=edges)
    hist.values = hist.values.astype(int)

    nonzero_counts = sparse_hist.values
    nonzero_indices = sparse_hist.indices

    assert len(nonzero_counts) == np.count_nonzero(hist.values)

    for idx, count in zip(nonzero_indices, nonzero_counts):
        assert hist.values[tuple(idx)] == count


def test_get_radii():
    x = gen_dist(ndim=6)
    radii = ps.get_radii(x)
    assert np.all(radii >= 0)
    cov_matrix = np.cov(x.T)
    radii = ps.get_radii(x, cov_matrix)
    assert np.all(radii >= 0)


def test_enclosing_sphere_radius():
    x = gen_dist(ndim=6)
    radius = ps.enclosing_sphere_radius(x)
    assert radius > 0


def test_enclosing_ellipsoid_radius():
    x = gen_dist(ndim=6)
    radius = ps.enclosing_ellipsoid_radius(x)
    assert radius > 0


def test_limits():
    x = gen_dist(ndim=6)
    limits = ps.limits(x)
    limits = ps.limits(x, rms=1.0)
    limits = ps.limits(x, share=(0, 1))
    limits = ps.limits(x, zero_center=True)


def test_project():
    x = gen_dist(ndim=6)
    x_proj = ps.project(x, axis=(0, 1))
    assert np.all(np.equal(x_proj, x[:, (0, 1)]))


def test_transform_linear():
    x = gen_dist(ndim=6)
    matrix = np.random.rand(6, 6)
    y = ps.transform_linear(x, matrix)
    assert y.shape == x.shape


def test_slice():
    x = gen_dist(ndim=6)
    x_slice = ps.slice_(x, axis=(2, 3), center=0.5, width=0.2)
    assert x_slice.shape[1] == 6


def test_slice_sphere():
    x = gen_dist(ndim=6)
    x_slice = ps.slice_sphere(x, rmax=1.0)
    assert np.all(ps.get_radii(x_slice) <= 1.0)


def test_slice_ellipsoid():
    x = gen_dist(ndim=6)
    x_slice = ps.slice_ellipsoid(x, rmax=1.0)


def test_slice_contour():
    x = gen_dist(ndim=2)
    ps.slice_contour(x, lmax=0.5)


def test_normalize_2d_projections():
    x = gen_dist(ndim=6)
    x_new = ps.normalize_2d_projections(x, scale=True)
    assert x.shape == x_new.shape


def test_decorrelate_x_y_z():
    x = gen_dist(ndim=6)
    x_new = ps.decorrelate_x_y_z(x)
    assert x.shape == x_new.shape


def test_downsample():
    x = gen_dist(ndim=6)
    x_new = ps.downsample(x, size=50)
    assert x_new.shape == (50, 6)


def test_histogram_bin_edges():
    x = gen_dist(ndim=2)
    edges = ps.histogram_bin_edges(x, bins=10)
    assert len(edges) == 2


def test_histogram():
    x = gen_dist(ndim=2)
    hist = ps.histogram(x, bins=10)
    assert hist.ndim == 2


def test_sparse_histogram():
    x = gen_dist(ndim=2)
    sparse_hist = ps.sparse_histogram(x, bins=10)
    assert hasattr(sparse_hist, "values")


def test_radial_histogram():
    x = gen_dist(ndim=2)
    hist = ps.radial_histogram(x, bins=10)


def test_density_estimation_gaussian_kde():
    x = gen_dist(ndim=2)
    x_eval = x

    estimator = ps.GaussianKDE(x)
    density = estimator(x_eval)
    assert len(density) == x.shape[0]
